/* Interhack   Multi-player Rougelike game (based on Nethack)
   Copyright (C) 1996-2000  Damian Bentley

   See license.txt for details. If you do not have a copy of the license,
   please email bendchon@mehta.anu.edu.au */

#include "branch.h"
#include "log.h"
#include "dungeon.h"
#include "string.h"

CBranch::CBranch(CBranch *n, CBranch *p, char *str, int l) {
    name = str;
    next = n;
    branches = NULL;
    levels = NULL;
    parent = p;
    id[0] = '\0';
    branchType = normalBranch;
    alignment = A_NONE;
    hellish = FALSE;
    depth = 0;
    location = l;
    entryType = EEntryDown;
    noDown = false;
}

CBranch::~CBranch() {
    if (next != NULL)
        delete next;
    next = NULL;
    if (branches != NULL)
        delete branches;
    branches = NULL;
    if (levels != NULL)
        delete levels;
    levels = NULL;
}

void CBranch::create() {
    /* Create the last branch in the list first (order of dungeon.def
       file is important */
    entryType = EEntryDown;
    if (next != NULL)
        next->create();

    /* Check for the right dungeon entry. */
    char * entry = deffile.readline();
    if (entry == NULL) {
        log << "ERROR: no branch " << name
             << " defined in " << DEFFILE << '\n';
        exit(-1);
    }
    if (strcmp(deffile.linetype(), "DUNGEON") != 0) {
        log << "ERROR: expected DUNGEON: " << name
             << " found " << entry << '\n';
        exit(-1);
    }
    string temp = deffile.first_field();
    if (temp != name) {
        log << "ERROR: expected DUNGEON: " << name
             << " found " << entry << '\n';
        exit(-1);
    }

    /* Configure this branch */
    char * flags;
    flags = deffile.next_field();
    id[0] = flags[0]; id[1] = flags[1]; id[2] = '\0';
    depth = getRand(deffile.next_field());
    for (flags = deffile.next_field(); flags != NULL;
        flags = deffile.next_field()) {
        if (strcmp(flags, "mazelike") == 0)
            branchType = mazeBranch;
        else if (strcmp(flags, "minelike") == 0)
            branchType = mineBranch;
        else if (strcmp(flags, "sewerlike") == 0)
            branchType = sewerBranch;
        else if (strcmp(flags, "unaligned") == 0)
            alignment = A_NONE;
        else if (strcmp(flags, "lawful") == 0)
            alignment = A_LAWFUL;
        else if (strcmp(flags, "noalign") == 0)
            alignment = A_NONE;
        else if (strcmp(flags, "chaotic") == 0)
            alignment = A_CHAOTIC;
        else if (strcmp(flags, "neutral") == 0)
            alignment = A_NEUTRAL;
        else if (strcmp(flags, "hellish") == 0)
            hellish = TRUE;
        else if (strcmp(flags, "portal") == 0)
            entryType = EEntryPortal;
        else if (strcmp(flags, "up") == 0)
            entryType = EEntryUp;
        else if (strcmp(flags, "no_down") == 0)
            noDown = true;
        else
            log << "WARNING: Unknown flag: " << flags << '\n';
    }

    /* Loop through all definition entries until ENDDUNGEON is found */
    while (1) {
        entry = deffile.readline();
        if (entry == NULL) {
            log << "ERROR: unexpected eof in DUNGEON: " << name << '\n';
            exit(-1);
        }
        entry = deffile.linetype();

        if (strcmp(entry, "ENDDUNGEON") == 0)
            break;

        else if (strcmp(entry, "BRANCH") == 0) {
            deffile.n_field(2);
            int lnum = deffile.read_random();
            if (lnum == -333) {
                log << "ERROR: branch location invalid, found " << temp << "\n";
                exit(-1);
            }
            branches = new CBranch(branches, this, deffile.first_field(), lnum);

        } else if (strcmp(entry, "CHAINBRANCH") == 0) {
            string temp;
            temp = deffile.n_field(2);
            int lnum = deffile.read_random();
            if (lnum == -333) {
                log << "ERROR: chainbranch location invalid, found " << temp << "\n";
                exit(-1);
            }
            char * tempbase = deffile.read_base();
            if (tempbase == NULL) {
                log << "ERROR: no base, found " << temp << "\n";
                exit(-1);
            }
            int lbase = dungeon->getNum(tempbase);
            if (lbase == -333) {
                log << "ERROR: chainbranch base invalid, found " << temp << "\n";
                exit(-1);
            }
            lnum = lbase + lnum;
            branches = new CBranch(branches, this, deffile.first_field(), lnum);

        } else if (strcmp(entry, "TEMPLATE") == 0) {
            int lnum = 0, choices = 1;
            string name, desfile, idletter;
            string temp;
            name = deffile.n_field(1);
            desfile = deffile.n_field(2);
            idletter = deffile.n_field(3);
            /* Extract the level number of this branch */
            temp = deffile.n_field(4);
            lnum = deffile.read_random();
            if (lnum == -333) {
                log << "ERROR: Expected \"@ (X, Y)\", found " << temp << "\n";
                exit(-1);
            }
            if (lnum < 0) lnum = depth + lnum;
            if (lnum < 0) {
                log << "ERROR: template level < 0, found " << temp << "\n";
                exit(-1);
            }

            deffile.n_field(5);
            choices = deffile.read_number();
            levels = new CTemplateLevel(levels, this, lnum, name,
                                        desfile, idletter, choices);
            levels->addUp();
            levels->addDown();

        } else if (strcmp(entry, "CHAINTEMPLATE") == 0) {
            int lnum = 0, lbase = 0, choices = 1;
            string name, desfile, idletter;
            string temp;
            name = deffile.n_field(1);
            desfile = deffile.n_field(2);
            idletter = deffile.n_field(3);
            /* Extract the level number of this branch */
            temp = deffile.n_field(4);
            lnum = deffile.read_random();
            if (lnum == -333) {
                log << "ERROR: Expected \"+ (X, Y)\", found " << temp << "\n";
                exit(-1);
            }
            if (lnum < 0) lnum = depth + lnum;
            if (lnum < 0) {
                log << "ERROR: template level < 0, found " << temp << "\n";
                exit(-1);
            }
            char * tempbase = deffile.read_base();
            if (tempbase == NULL) {
                log << "ERROR: expected \"base\" + (X, Y), found " << temp << "\n";
                exit(-1);
            }
            lbase = dungeon->getNum(tempbase);
            if (lbase == -333) {
                log << "ERROR: invalid base, found " << temp << "\n";
                exit(-1);
            }
            lnum = lbase + lnum;

            deffile.n_field(5);
            choices = deffile.read_number();
            levels = new CTemplateLevel(levels, this, lnum, name,
                                        desfile, idletter, choices);
            levels->addUp();
            levels->addDown();

        } else
            log << "WARNING: ignoring unknown line type: "
                 << entry << '\n';
    }

    /* Create all sub-branches in this one */
    if (branches != NULL)
        branches->create();
}

/* Get level n of this branch */
CLevel * CBranch::getLevel(int n) {
    if (n < 1) {
        log << "ERROR: creating level with number < 1\n";
        exit(-1);
    }

    if (n > depth) {
        log << "WARNING: creating level with number > branch depth\n";
        return NULL;
    }

    /* See if level is already created */
    CLevel * temp = NULL;
    if (levels)
        temp = levels->get(n);

    /* If level not already created, create a new one */
    if (temp == NULL) {
        switch (branchType) {
            case normalBranch:
                levels = new CNormalLevel(levels, this, n);
                break;
            case mineBranch:
                levels = new CMineLevel(levels, this, n);
                break;
            case mazeBranch:
                levels = new CMazeLevel(levels, this, n);
                break;
            case sewerBranch:
                levels = new CSewerLevel(levels, this, n);
                break;
            case templateBranch:
                log << "ERROR: Cannot create a level on a template branch\n";
                exit(-1);
                break;
            default:
                log << "ERROR: Unknown branch type. Cannot create level\n";
                exit(-1);
        }
        /* This should be temp = levels->get(n);, but it works... */
        temp = levels;

        /* Add the up and down stairs */
        if (n > 1)
            temp->addUp();
        temp->addDown();

        /* Add any branch */
        CBranch *tempb = NULL;
        CLocation * temploc = NULL;
        for (tempb = branches; tempb != NULL; tempb = tempb->next)
            if (tempb->location == n) {
                temploc = temp->addBranch(tempb->entryType);
                break;
            }

        /* Add entry if this is not the main branch */
        if (n == 1 && temploc == NULL) {
            if (entryType == EEntryDown)
                temp->addBranch(EEntryUp);
            else if (entryType == EEntryUp)
                temp->addBranch(EEntryDown);
            else
                temp->addBranch(entryType);
        }

        temp->populate();
    }

    /* Return the client map */
    return temp;
}

/* Returns level 1 of a branch on level n of the current branch...
   I know it sounds bad, and there should be a better way, but...
   Even the error message is cryptic! */
CLevel * CBranch::getBranchLevel1(int n) {
    /* Find the branch on level n */
    CBranch * tempb = NULL;
    CLevel * templ = NULL;
    for (tempb = branches; tempb != NULL; tempb = tempb->next)
        if (tempb->location == n) break;

    /* Get the first level on that branch */
    if (tempb != NULL)
        templ = tempb->getLevel(1);

    if (templ == NULL)
        log << "WARNING: could not find branch on level " << n
             << " to get first level\n";

    return templ;
}

/* Returns a number based on a string "(X Y)": return X + rand%Y; */
int CBranch::getRand(char *str) {
    int X = 0, Y = 0, i;
    char tmp[10], *tmpend;

    /* Find first number (X) */
    mystrcpy(tmp, str+1);
    for (tmpend = tmp; *tmpend != ','; tmpend++) ;
    *tmpend = '\0';
    for (i = 0; i < (signed)mystrlen(tmp); i++)
        X = (X * 10) + (tmp[i] - '0');

    /* Find second number (Y) */
    mystrcpy(tmp, str+i+3);
    for (tmpend = tmp; *tmpend != ')'; tmpend++) ;
    *tmpend = '\0';
    for (i = 0; i < (signed)mystrlen(tmp); i++)
        Y = (Y * 10) + (tmp[i] - '0');

    /* Return random number */
    if (Y > 0)
        return X + (rn2(Y));
    return X;
}

/* Return the level number of level or branch with idstr as name */
int CBranch::getNum(char *idstr) {
    int n = -333;
    if (levels != NULL) n = levels->getNum(idstr);
    if (branches != NULL && n < 0) n = branches->getNum(idstr);
    if (next != NULL && n < 0) n = next->getNum(idstr);
    return n;
}

/* Get a level based on its branch name and number */
CLevel * CBranch::get(string s) {
    CLevel * tlvl = NULL;
    string tstr;
    char * c;
    /* If the branch is this one, get the level */
    if (name.size() < s.size() && mystrncmp(name.getStr(), s.getStr(), name.size()) == 0) {
        for (c = s.getStr(); c < (s.getStr() + s.size()); c++)
            if (*c >= '0' && *c <= '9') break;
        if (*c >= '0' && *c <= '9') {
            tstr = c;
            tlvl = getLevel(tstr.getNumber());
        }
    }
    /* If there are other branches linked to this one, check them */
    if (tlvl == NULL && branches != NULL) tlvl = branches->get(s);
    if (tlvl == NULL && next != NULL)     tlvl = next->get(s);
    return tlvl;
}
