/* Interhack   Multi-player Rougelike game (based on Nethack)
   Copyright (C) 1996-2000  Damian Bentley

   See license.txt for details. If you do not have a copy of the license,
   please email bendchon@mehta.anu.edu.au */

#include "level.h"
#include "log.h"
#include "branch.h"
#include "vision.h"

/* Send the list to the player */
void CLevel::send(CSock *s, ID user) {
    for (register int y = 0; y < HEIGHT; y++)
        for (register int x = 0; x < WIDTH; x++)
            *s << (char)(map[y][x]);
    objects.send(s, user);   /* Send objects on the level */
    *s << monsters; /* Send monsters on the level */
}

// Setup the initial level stats and details
CLevel::CLevel(CLevel *newNext, CBranch *newParent, int newNumber) {
    next = newNext;
    parent = newParent;
    number = newNumber;
    map = NULL;
    map = new EFloorType*[HEIGHT];
    if (map == NULL)
        log << "ERROR CLevel::CLevel(CLevel *, int) - not enough memory\n";
    for (int y = 0; y < HEIGHT; y++) {
        map[y] = NULL;
        map[y] = new EFloorType[WIDTH];
        if (map[y] == NULL)
            log << "ERROR CLevel::CLevel(CLevel *, int) - not enough memory\n";
    }
    isTempCopy = 0;
    clean();
    rooms = NULL;
}

// Populate the dungeon with monsters and objects
void CLevel::populate() {
    if (isTemplate() == true)
        return;
    objects.createLevel(this);
    monsters.createLevel(this);
}

// Clean up this level and all children below it
CLevel::~CLevel() {
    if (next != NULL && isTempCopy == 0) {
        delete next;
        next = NULL;
    }
    if (map != NULL && isTempCopy == 0) {
        for (int y = 0; y < HEIGHT; y++)
            if (map[y] != NULL) {
                delete [] map[y];
                map[y] = NULL;
            }
        delete [] map;
        map = NULL;
    }
    if (rooms != NULL && isTempCopy == 0) {
        delete rooms;
        rooms = NULL;
    }
}

// Create a temporary copy of a level
// Instead of creating a completely new copy, we simply copy the vital
// information, and make a copy of the pointers. When we come to delete
// a temporary copy of the object, do not delete the pointers.
CLevel::CLevel(CLevel& l) {
    isTempCopy = 1;
    number = l.number;
    next = l.next;
    map = l.map;
    objects = l.objects;
	monsters = l.monsters;
    rooms = l.rooms;
}

// Load a level from file returns -1 on error
int CLevel::load() {
    // If the map is already in memory do not need to load it in
    if (map != NULL)
        return 0;

    // Open the file
    char filename[] = "fred";
    ifstream inFile(filename, ios::in);
    if (!inFile) {
        log << "ERROR: Clevel::load - could not open file to load level\n";
        return -1;
    }

    // Read in the level
    map = new EFloorType*[HEIGHT];
    if (map == NULL)
        log << "ERROR CLevel::load() - not enough memory\n";
    for (int j = 0; j < HEIGHT; j++) {
        map[j] = NULL;
        map[j] = new EFloorType[WIDTH];
        if (map[j] == NULL)
            log << "ERROR CLevel::load() - not enough memory\n";
        for (int i = 0; i < WIDTH; i++) {
            char ch;
            inFile.get(ch);
            map[j][i] = (EFloorType)ch;
        }
    }

    // Load all the objects on the level to the file
    log << "FIXME: CLevel::load - Load objects from file\n";

    // Close the file
    inFile.close();

    // Delete the save file
    log << "FIXME: CLevel::load - Delete the save file\n";

    return 0;
}

// Save a level to file returns -1 on error
int CLevel::save() {
    // If the map is already saved, do not worry about it
    if (map == NULL)
        return 0;

    // Create the new file
    char filename[] = "fred";
    ofstream outFile(filename, ios::out);
    if (!outFile) {
        log << "ERROR: Clevel::load - could not open file to save level\n";
        return -1;
    }

    // Save the level information to the file
    // At a later point in time compression might be used here
    for (register int j = 0; j < HEIGHT; j++) {
        for (register int i = 0; i < WIDTH; i++)
            outFile.put((char)map[j][i]);
        delete [] map[j];
        map[j] = NULL;
    }
    delete [] map;
    map = NULL;

    // Save all the objects on the level to the file
    log << "FIXME: CLevel::save - Save objects to file\n";

    // Close the file
    outFile.close();

    return 0;
}

chtype sfloorSym[] = {
    ' ', ' ', '.', ACS_BOARD,
    ACS_HLINE, ACS_VLINE, ACS_ULCORNER, ACS_URCORNER, ACS_LLCORNER, ACS_LRCORNER,
    ACS_PLUS, ACS_BTEE, ACS_TTEE, ACS_RTEE, ACS_LTEE,
    '~', '~', '<', '>', '.'};

/* Display a level (on the server side) */
void CLevel::display() {
    for (register int y = 0; y < HEIGHT; y++)
        for (register int x = 0; x < WIDTH; x++) {
            EFloorType t = map[y][x];
            if (t == EFTHWall) vision.mapxy(x, y, ACS_HLINE);
            else if (t == EFTVWall) vision.mapxy(x, y, ACS_VLINE);
            else if (t == EFTTLWall) vision.mapxy(x, y, ACS_ULCORNER);
            else if (t == EFTTRWall) vision.mapxy(x, y, ACS_URCORNER);
            else if (t == EFTBLWall) vision.mapxy(x, y, ACS_LLCORNER);
            else if (t == EFTBRWall) vision.mapxy(x, y, ACS_LRCORNER);
            else if (t == EFTCorridor) vision.mapxy(x, y, ACS_BOARD);
            else vision.mapxy(x, y, sfloorSym[t]);
        }
}

string CLevel::getName() {
    string s = parent->getName();
    if (s == "The Dungeons of Doom") {
        s = number; return s; };
    s += ' ';
    s += number;
    return s;
};

// Add a room at location x,y with width, height size
void CLevel::addRoom(uchar x, uchar y, uchar width, uchar height) {
    // Check to see the map is in memory
    if (map == NULL)
        if (load() == -1)
            return;

    // Place a floor everywhere, with walls on the side
    for (int j = y; j < y+height; j++)
        for (int i = x; i < x+width; i++) {
            if (i == x || i == x+width-1)
                map[j][i] = EFTVWall;
            else if (j == y || j == y+height-1)
                map[j][i] = EFTHWall;
            else
                map[j][i] = EFTFloor;
        }

    // Add the four corner walls
    map[y][x] = EFTTLWall;
    map[y][x+width-1] = EFTTRWall;
    map[y+height-1][x] = EFTBLWall;
    map[y+height-1][x+width-1] = EFTBRWall;

    // Add the room to temporary storage
    rooms = new CRoom(rooms, x, y, width, height);
}

// Add a sub-room at location x,y with width, height size
void CLevel::addSubRoom(uchar x, uchar y, uchar width, uchar height) {
    // Check to see the map is in memory
    if (map == NULL)
        if (load() == -1)
            return;

    // Place a floor everywhere, with walls on the side
    for (int j = y; j < y+height; j++)
        for (int i = x; i < x+width; i++) {
            if (i == x || i == x+width-1)
                map[j][i] = EFTVWall;
            else if (j == y || j == y+height-1)
                map[j][i] = EFTHWall;
            else
                map[j][i] = EFTFloor;
        }

    // Add the four corner walls
    map[y][x] = EFTTLWall;
    map[y][x+width-1] = EFTTRWall;
    map[y+height-1][x] = EFTBLWall;
    map[y+height-1][x+width-1] = EFTBRWall;
}

// Add a room within x,y,width,height
void CLevel::addRoomWithin(uchar x, uchar y, uchar width, uchar height) {
    int roomX, roomY, roomWidth, roomHeight;
    // Check to see the map is in memory
    if (map == NULL)
        if (load() == -1)
            return;

    /* Create a buffer zone around all rooms - mainly to allow corridors */
    x+=1; y+=1; width-=2; height-=2;

    /* Try to create a random room 1000 times, if that fails just dump the
       last room created */
    for (register int i = 0; i < 1000; i++) {

        /* Select a random width and height for the room */
        roomWidth = (rn2(MAXROOMWIDTH-MINROOMWIDTH)) + MINROOMWIDTH;
        roomHeight = (rn2(MAXROOMHEIGHT-MINROOMHEIGHT)) + MINROOMHEIGHT;
        if (width < roomWidth) roomWidth = width;
        if (height < roomHeight) roomHeight = height;

        /* Select a random location for the room */
        roomX = x + (rn2(width - roomWidth + 1));
        roomY = y + (rn2(height - roomHeight + 1));

        if (rooms != NULL &&
            !rooms->overlapping(roomX-1, roomY-1, roomWidth+1, roomHeight+1))
            break;
    }

    addRoom(roomX, roomY, roomWidth, roomHeight);
}

//ellipse function.  MAX SIZE IS WIDTH = 10, HEIGHT = 10
void CLevel::ellipse(int width, int height, int placey, int placex) {
    //We must manipulate it character wise until we are ready to port it over
    EFloorType ellipsemap[20][20];
    //initialize
    for (int tempy=0;tempy<20;tempy++)
        for(int tempx=0;tempx<20;tempx++)
            ellipsemap[tempy][tempx] = EFTEmpty;

    //make a quarter of the curve of the ellipse
    for (float x=1;x<=width;x++) {
        float y;
        //THE ellipse definition, dont touch it or else it will explode :)
        y = sqrt(((((x*x)/(width*width))*-1)+1)*(height*height));
        ellipsemap[(20-(10-int(y)))][(20-(10-int(x)))] = EFTFloor;
    }
    //place the bottom middle edge, cannot do in previous loop due to division by 0
    ellipsemap[(20-(11-int(height)))][(20-(10))] = EFTFloor;

    //fill in the rest of the quarter circle
    for(int tempx=0;tempx<width;tempx++) {
       int tempy=0;
       do {
           if(ellipsemap[10+tempy][10+tempx] != EFTFloor)
               ellipsemap[10+tempy][10+tempx] = EFTFloor;
           tempy++;
       }while(ellipsemap[10+tempy][10+tempx] != EFTFloor);
    }

    //using the quarter circle, create the other 3 quadrants
    for(int tempy=0;tempy<10;tempy++) {
        for(int tempx=0;tempx<10;tempx++) {
            if(ellipsemap[(20-(10-int(tempy)))][(20-(10-int(tempx)))] == EFTFloor) {
                ellipsemap[(0+(10-int(tempy)))][(20-(10-int(tempx)))] = EFTFloor;
                ellipsemap[(20-(10-int(tempy)))][(0+(10-int(tempx)))] = EFTFloor;
                ellipsemap[(0+(10-int(tempy)))][(0+(10-int(tempx)))] = EFTFloor;
            }
        }
    }

    //dump it to the map
    for(int tempy=0;tempy<20;tempy++)
        for(int tempx=0;tempx<20;tempx++)
            if(ellipsemap[tempy][tempx] == EFTFloor)
                map[(placey+(tempy-10))][(placex+(tempx-8))] = ellipsemap[tempy][tempx];
}

//solidline - makes a line that is never connected by diagonals
void CLevel::solidline(int point1y, int point1x, int point2y, int point2x) {
    //for loops and ratios
    float diffy = abs(int(point2y) - int(point1y));
    if (diffy == 0) diffy = 1;

    float diffx = abs(int(point2x) - int(point1x));
    if (diffx == 0) diffx = 1;
    int lasty = -1,lastx = -1;
    //if x is longer than y
    if(abs(int(diffx)/int(diffy)) >= 1) {
       
        //switch the 2 points for the loop
        if(point2y < point1y) {
	    int temp = point2y;
            point2y = point1y;
            point1y = temp;
          
            temp = point2x;
            point2x = point1x;
            point1x = temp;
        }

        //plot each corridor slot
        for(int x = 0 ; x <= diffx; x++) {
            int tempy = int((((x/diffx)*diffy)+point1y));
            int tempx;
            if(point1x > point2x)
                 tempx = int((-1*x)+int(point1x));
            else
       	         tempx = int(x+int(point1x));
              
            if(tempx > 79) tempx = 79;
            if(tempy > 19) tempy = 19;
            if(tempx < 0) tempx = 0;
            if(tempy < 0) tempy = 0;
	   
            //the diagonal connector
	    if(lasty != tempy && lasty != -1 && map[lasty][tempx] != EFTWater &&
	       map[lasty][tempx] != EFTFloor)
	         map[lasty][tempx] = EFTCorridor;
		
            //regular plot
            if(map[tempy][tempx] != EFTWater && map[tempy][tempx] != EFTFloor)
	        map[tempy][tempx] = EFTCorridor;
          
            lasty = tempy;
            lastx = tempx;
        }
    }
    //if y is longer than x
    //all comments are the same as above, but x and y are switched
    else {
        if(point2x < point1x) {
            int temp = point2y;
            point2y = point1y;
            point1y = temp;
          
            temp = point2x;
            point2x = point1x;
            point1x = temp;
        }
        for(int y = 0 ; y <= diffy; y++) {
            int tempx = int((((y/diffy)*diffx)+point1x));
            int tempy;
            if(point1y > point2y)
                tempy =  int((-1*y)+int(point1y));    
            else
       	        tempy =  int(y+int(point1y));    
          
            if(tempx > 79) tempx = 79;
            if(tempy > 19) tempy = 19;
            if(tempx < 0) tempx = 0;
            if(tempy < 0) tempy = 0;
      
	    if(lastx != tempx && lastx != -1 && map[tempy][lastx] != EFTWater &&
              map[tempy][lastx] != EFTFloor)
                map[tempy][lastx] = EFTCorridor;
	      
            if(map[tempy][tempx] != EFTWater && map[tempy][tempx] != EFTFloor)
                map[tempy][tempx] = EFTCorridor;
            
            lasty = tempy;
            lastx = tempx;
        }
    }
}
    
// Clean the map
void CLevel::clean() {
    // Check to see the map is in memory
    if (map == NULL)
        if (load() == -1)
            return;

    for (register int j = 0; j < HEIGHT; j++)
        for (register int i = 0; i < WIDTH; i++)
            map[j][i] = EFTEmpty;
}

// Return the pointer to level n
CLevel * CLevel::get(int n) {
    if (n == number)
        return this;
    if (next != NULL)
        return next->get(n);
    return NULL;
}

/* Returns the level from which the current location loc on lvl links to.
   NULL if loc does not link anywhere */
CLevel * CLevel::get(CLocation loc) {
    if (loc == branch) {
        /* If anyone wants to sort this one out they can... yuck... */
        if (number == 1 && parent->getParent() != NULL)
            return parent->getParent()->getLevel(parent->getLocation());
        return parent->getBranchLevel1(number);
    } else if (loc == upStair) {
        if (number == 1) {
            log << "WARNING: player going up stairs when this is level 1\n";
            return NULL;
        }
        return parent->getLevel(number - 1);
    } else if (loc == downStair)
        return parent->getLevel(number + 1);
    return NULL;
}


/* Add some up stairs */
CLocation * CLevel::addUp() {
    do
        upStair = CLocation(rn2(80), rn2(20));
    while (map[upStair.y][upStair.x] != EFTFloor);
    map[upStair.y][upStair.x] = EFTUpStair;
    return &upStair;
}

/* Add some down stairs */
CLocation * CLevel::addDown() {
    do
        downStair = CLocation(rn2(80), rn2(20));
    while (map[downStair.y][downStair.x] != EFTFloor);
    map[downStair.y][downStair.x] = EFTDownStair;
    return &downStair;
}

/* Add some the branch */
CLocation * CLevel::addBranch(EEntryType entryType) {
    do
        branch = CLocation(rn2(80), rn2(20));
    while (map[branch.y][branch.x] != EFTFloor);
    switch (entryType) {
        case EEntryPortal:
            map[branch.y][branch.x] = EFTPortal; break;
        case EEntryDown:
            map[branch.y][branch.x] = EFTDownStair; break;
        case EEntryUp:
            map[branch.y][branch.x] = EFTUpStair; break;
        default:
            log << "WARNING: Unknown branch entry type.\n";
    }
    return &branch;
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
/* Quadrant I - Checks the path from sx,sy to fx, fy,
   returns 1 if path is blocked, 0 otherwise. It does not check fx, fy; */
#define blocked1(sx, sy, fx, fy) {				\
    int dx, dy;							\
    register int k, err, x, y, dxs, dys;			\
								\
    x = sx;		y = sy;					\
    dx = fx - x;	dy = y - fy;				\
								\
    dxs = dx << 1;						\
    dys = dy << 1;						\
    if (dy > dx) {						\
        err = dxs - dy;						\
        for (k = dy - 1; k; k--) {				\
            if (err >= 0) {					\
                x++;						\
                err -= dys;					\
            }							\
            y--;						\
            err += dxs;						\
            last_lit = lit(x, y);				\
            if (blocked(x,y))					\
                return 1;					\
        }							\
    } else {							\
        err = dys - dx;						\
        for (k = dx - 1; k; k--) {				\
            if (err >= 0) {					\
                y--;						\
                err -= dxs;					\
            }							\
            x++;						\
            err += dys;						\
            last_lit = lit(x, y);				\
            if (blocked(x, y))					\
                return 1;					\
        }							\
    }								\
    return 0;							\
}								\

/* Quadrant IV - Checks the path from sx,sy to fx, fy,
   returns 1 if path is blocked, 0 otherwise. It does not check fx, fy; */
#define blocked4(sx, sy, fx, fy) {				\
    int dx, dy;							\
    register int k, err, x, y, dxs, dys;			\
								\
    x  = sx;		y  = sy;				\
    dx = fx - x;	dy = fy - y;				\
								\
    dxs = dx << 1;						\
    dys = dy << 1;						\
    if (dy > dx) {						\
        err = dxs - dy;						\
        for (k = dy-1; k; k--) {				\
            if (err >= 0) {					\
                x++;						\
                err -= dys;					\
            }							\
            y++;						\
            err += dxs;						\
            last_lit = lit(x, y);				\
            if (blocked(x,y))					\
                return 1;					\
        }							\
    } else {							\
        err = dys - dx;						\
								\
        for (k = dx-1; k; k--) {				\
            if (err >= 0) {					\
                y++;						\
                err -= dxs;					\
            }							\
            x++;						\
            err += dys;						\
            last_lit = lit(x, y);				\
            if (blocked(x,y))					\
                return 1;					\
        }							\
    }								\
    return 0;							\
}								\

/* Quadrant II - Checks the path from sx,sy to fx, fy,
   returns 1 if path is blocked, 0 otherwise. It does not check fx, fy; */
#define blocked2(sx, sy, fx, fy) {				\
    int dx, dy;							\
    register int k, err, x, y, dxs, dys;			\
								\
    x  = sx;		y  = sy;				\
    dx = x - fx;	dy = y - fy;				\
								\
    dxs = dx << 1;						\
    dys = dy << 1;						\
    if (dy > dx) {						\
        err = dxs - dy;						\
        for (k = dy-1; k; k--) {				\
            if (err >= 0) {					\
                x--;						\
                err -= dys;					\
            }							\
            y--;						\
            err += dxs;						\
            last_lit = lit(x, y);				\
            if (blocked(x, y))					\
                return 1;					\
        }							\
    } else {							\
        err = dys - dx;						\
        for (k = dx-1; k; k--) {				\
            if (err >= 0) {					\
                y--;						\
                err -= dxs;					\
            }							\
            x--;						\
            err += dys;						\
            last_lit = lit(x, y);				\
            if (blocked(x, y))					\
                return 1;					\
        }							\
    }								\
    return 0;							\
}								\

/* Quadrant III - Checks the path from sx,sy to fx, fy,
   returns 1 if path is blocked, 0 otherwise. It does not check fx, fy; */
#define blocked3(sx, sy, fx, fy) {				\
    int dx, dy;							\
    register int k, err, x, y, dxs, dys;			\
								\
    x  = sx;		y  = sy;				\
    dx = x - fx;	dy = fy - y;				\
								\
    dxs = dx << 1;						\
    dys = dy << 1;						\
    if (dy > dx) {						\
        err = dxs - dy;						\
        for (k = dy-1; k; k--) {				\
            if (err >= 0) {					\
                x--;						\
                err -= dys;					\
            }							\
            y++;						\
            err += dxs;						\
            last_lit = lit(x, y);				\
            if (blocked(x, y))					\
                return 1;					\
        }							\
    } else {							\
        err = dys - dx;						\
        for (k = dx-1; k; k--) {				\
            if (err >= 0) {					\
                y++;						\
                err -= dxs;					\
            }							\
            x--;						\
            err += dys;						\
            last_lit = lit(x, y);				\
            if (blocked(x, y))					\
                return 1;					\
        }							\
    }								\
    return 0;							\
}								\

/* This decides which quadrant to use,
   returns 1 if path is blocked, 0 otherwise */
int CLevel::blocked_path(int sx, int sy, int fx, int fy) {
    if (sx < fx) {
        if (sy > fy) {
            blocked1(sx, sy, fx, fy);
        } else {
            blocked4(sx, sy, fx, fy);
        }
    } else {
        if (sy > fy) {
            blocked2(sx, sy, fx, fy);
        } else if (sx == fx && sy == fy) {
            return 1;
        } else {
            blocked3(sx, sy, fx, fy);
        }
    }
}


/* Returns 1 if x,y is lit, 0 otherwise */
int CLevel::lit(int x, int y) {
    if (x < 1 || x >= WIDTH || y < 1 || y > HEIGHT)
        return 0;
    /* For the time being this is not done on a proper basis... */
    /* log << "FIXME: Floor lighting in CClientLevel::lit\n"; */
    if (map[y][x] == EFTFloor)
        return 1;
    return 0;
}

/* Returns 1 if there is a blocking object in location (x, y),
   0 otherwise */
int CLevel::blocked(int x, int y) {
    if (x < 1 || x >= WIDTH || y < 1 || y >= HEIGHT)
        return 1;
    EFloorType t = map[y][x];
    if ((t >= EFTHWall && t <= EFTWTWall) || t == EFTEmpty)
        return 1;
    return 0;
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/* Get the starting location of a player based on their current location */
CLocation CLevel::getStart(CLocation xy) {
    if (xy == upStair && number > 1)
        return (parent->getLevel(number-1))->getDown();
    if (xy == downStair)
        return (parent->getLevel(number+1))->getUp();
    else if (xy == branch) {
        /* If anyone wants to sort this one out they can... yuck... */
        if (number == 1 && parent->getParent() != NULL)
            return parent->getParent()->getLevel(parent->getLocation())->getBranch();
        return parent->getBranchLevel1(number)->getBranch();
    }
    CLocation temp(-1,-1);
    return temp;
}

/* Get a random location on the level */
CLocation CLevel::getRandomFloor() {
    int x, y;
    do {
        x = rand() % WIDTH;
        y = rand() % HEIGHT;
    } while (map[y][x] != EFTFloor);
    CLocation l(x, y);
    return l;
}

CLocation CLevel::getRandomCorridor() {
    int x, y;
    do {
        x = rand() % WIDTH;
        y = rand() % HEIGHT;
    } while (map[y][x] != EFTCorridor);
    CLocation l(x, y);
    return l;
}

CLocation CLevel::getRandomAny() {
    int x, y, count = 0;
    while (count++ < 2000) {
        x = rand() % WIDTH;
        y = rand() % HEIGHT;
        if (map[y][x] == EFTFloor || map[y][x] == EFTCorridor) break;
    }
    if (count >= 2000) {
        for (y = 0; y < HEIGHT; y++) {
            for (x = 0; x < WIDTH; x++)
                if (map[y][x] == EFTFloor || map[y][x] == EFTCorridor) break;
            if (map[y][x] == EFTFloor || map[y][x] == EFTCorridor) break;
        }
        if (x == WIDTH && y == HEIGHT) { x = y = 10; };
    }
    CLocation l(x, y);
    return l;
}

////////////////////////////////////////////////////////////////////////////////

/* Create coridors to each room */
void CLevel::addCorridors() {
    /* Go through each room and connect it to the next in the list 
       this makes sure _all_ rooms are connected */
    CRoom * from;
    CLocation * dfrom, * dto, * temp;
    for (from = rooms; from->getNext() != NULL; from = from->getNext()) {
        dfrom = from->getADoor();
        temp = from->getTheDoor();
        map[temp->y][temp->x] = EFTFloor;
        dto = from->getNext()->getADoor();
        temp = from->getNext()->getTheDoor();
        map[temp->y][temp->x] = EFTFloor;
        doCorridor(*dfrom, *dto);
    }
}

void CLevel::doCorridor(CLocation start, CLocation finish) {
    register int sx, sy, fx, fy;
    register int ox, oy, odd = 0;
    int overflow = 0;

    /* Get the start and finish points of the coridor */
    sx = start.x;
    sy = start.y;
    fx = finish.x;
    fy = finish.y;
    map[sy][sx] = EFTCorridor;
    map[fy][fx] = EFTCorridor;

    /* This is the loop to create the coridor */
    while (sx != fx || sy != fy || sx < 0 || sy < 0 || fx < 0 || fy < 0) {
        if (++overflow > 2000) break;
        ox = sx;
        oy = sy;
        /* Do any diagonal sections first */
        if (sx != fx && sy != fy) {
            odd++;
            if (odd == 1 || odd == 2) {
                if (sx < fx)       sx++;
                else if (sx > fx)  sx--;
            } else if (odd == 3 || odd == 4) {
                if (sy < fy)       sy++;
                else if (sy > fy)  sy--;
            }
            if (odd == 4)
                odd = 0;
        /* And now do the stright sections */
        } else if (sx != fx || sy != fy) {
            if (sx == fx) {
                if (sy < fy)       sy++;
                else               sy--;
            } else if (sy == fy) {
                if (sx < fx)       sx++;
                else               sx--;
            }
        }
        if (map[sy][sx] == EFTEmpty)
            map[sy][sx] = EFTCorridor;
        else if (map[sy][sx] != EFTCorridor)
            around(&sx, &sy, fx, fy, ox, oy);
    }
}

void CLevel::around(int *sx, int *sy, int fx, int fy, int ox, int oy) {
    register int x, y;
    register int d = 0; /* N=1, NE=2, E=3, SE=4, S=5, SW=6, W=7, NW=8 direction */
    register int s = 0; /* N=1, E=2, S=3, W=4  side of room */

    /* Find out where about we want to get to */
    if (ox == fx && oy > fy) d = 1;
    if (ox < fx && oy > fy)  d = 2;
    if (ox < fx && oy == fy) d = 3;
    if (ox < fx && oy < fy)  d = 4;
    if (ox == fx && oy < fy) d = 5;
    if (ox > fx && oy < fy)  d = 6;
    if (ox > fx && oy == fy) d = 7;
    if (ox > fx && oy > fy)  d = 8;

    /* Find out which side of the room we are on */
    if (oy < (*sy)) s = 1;
    if (ox > (*sx)) s = 2;
    if (oy > (*sy)) s = 3;
    if (ox < (*sx)) s = 4;

    /* These all deal with going round a room when at diagonals */
    if ((d == 4 && s == 4) || (d == 6 && s == 2)) {
        (*sx) = ox;
        (*sy) = oy + 1;
    } else if ((d == 2 && s == 4) || (d == 8 && s == 2)) {
        (*sx) = ox;
        (*sy) = oy - 1;
    } else if ((d == 6 && s == 1) || (d == 8 && s == 3)) {
        (*sx) = ox - 1;
        (*sy) = oy;
    } else if ((d == 4 && s == 1) || (d == 2 && s == 3)) {
        (*sx) = ox + 1;
        (*sy) = oy;
    /* The next two deal with strights going round rooms */
    } else if (d == 1 || d == 5) {
        int r = 0;
        /* Find the shortest way round */
        for (x = 1; x <= WIDTH; x++) {
            r = 1;
            if (ox + x < WIDTH-1) {
                if (map[(*sy)][ox + x] == EFTEmpty)    break;
                if (map[(*sy)][ox + x] == EFTCorridor) break;
            }
            r = -1;
            if (ox - x >= 1) {
                if (map[(*sy)][ox - x] == EFTEmpty)    break;
                if (map[(*sy)][ox - x] == EFTCorridor) break;
            }
        }
        /* Now create the coridor */
        for (y = ox; y < ox + x; y++) {
            (*sx) += r;
            if (*sx < 0 || *sx >= WIDTH) break;
            map[oy][(*sx)] = EFTCorridor;
        }
    } else if (d == 3 || d == 7) {
        int r = 0;
        /* Find the shortest way round */
        for (y = 1; y <= HEIGHT; y++) {
            r = 1;
            if (oy + y < HEIGHT-1) {
                if (map[oy + y][(*sx)] == EFTEmpty)    break;
                if (map[oy + y][(*sx)] == EFTCorridor) break;
            }
            r = -1;
            if (oy - y >= 1) {
                if (map[oy - y][(*sx)] == EFTEmpty)    break;
                if (map[oy - y][(*sx)] == EFTCorridor) break;
            }
        }
        /* Now create the coridor */
        for (x = oy; x < oy + y; x++) {
            (*sy) += r;
            if (*sy < 0 || *sy >= HEIGHT) break;
            map[(*sy)][ox] = EFTCorridor;
        }
    }

    if (map[(*sy)][(*sx)] == EFTEmpty)
        map[(*sy)][(*sx)] = EFTCorridor;
}
