/* Interhack   Multi-player Rougelike game (based on Nethack)
   Copyright (C) 1996-2000  Damian Bentley

   See license.txt for details. If you do not have a copy of the license,
   please email bendchon@mehta.anu.edu.au */

#include "level.h"
#include "log.h"

//This is the sewer level generation code.  For now this will be somewhat
//simple.  The genreal layout will look like this:
//
//       0-----------1
//        |v1  |v2  |v3
//       2-----------3
//        |    |    |
//       4-----------5
//
//   There will be 3+rand(3) of the sections numbred 1-6.  Each number
//represents half of a length of the dashes.  The vertical bars denote
//connections to those bars and are randomply generated depending on where
//the numbered sections have been defined.
//   The stairs going up and down will be randomly placed at one of the
//numbers above.  The entire sewer will always be dark.
//   The specific layout is the following:
//      ---------------------
//      .....................
//      WWWWWWWWWWWWWWWWWWWWW
//      .....................
//      ---------------------
//where the '.' is a floor, and 'W' is water.  So players wont have to
//always walk in water, but will have to at some points at intersections.
//   Also water monsters will be able to attack them and draw them not the water.
//   These levels will be somewhat dangerous.
//   This level will also help to put jumping to use my jumping over the water to
//get to the other side of a sewer passage.
//For now, 3+rand(2) levels like this will be genereated.
////////////////////////////////////////////////////////////////////////////////
CSewerLevel::CSewerLevel(CLevel *newNext, CBranch *newParent, int newNumber) :
    CLevel(newNext, newParent, newNumber) {
#ifdef DEBUG_LEVEL
    log << "LEVEL: constructing sewer " << newNumber << '\n';
#endif
    regenerate();
}

void CSewerLevel::regenerate() {
    // Clean the level up so there is no junk on it before regenerating.
    clean();

    //12 is how many places that the stairs can be generated.  Up, and Down
    //0           2
    //1           3

    //4           6
    //5           7

    //8           10
    //9           11
    int PosStairs[12];


    // Determine the number of levels
    //int nLevels = (rn2(2)) + 3;

    //Determine the numbers of horizontal sections
    int nSections = (rn2(3)) + 3;

    //Randomly place the sections
    int Placed[6] = {0,0,0,0,0,0};
    /* If after 500 times the sections aren't created, screw it :)
       ... hmmm bad idea... lets add a trick to ensure it works */
    for(int z = 0; z < 6; z++) {
        if (nSections == 0) break;
        int  randsec = (rn2(6));
        while (Placed[randsec] == 1)      /* Trick: If not this one, next */
            randsec = (randsec + 1) % 6;
        if (Placed[randsec] != 1) {
            Placed[randsec] = 1;
            nSections--;
        }
    }

////////////////////////////////////////////////////////////////////////////////

    //Create the sections
    int secstart = 0;
    for(int secloop = 0; secloop < 6; secloop++) {
        if (Placed[secloop] == 1) {
            PosStairs[secloop*2] = 1;
            PosStairs[(secloop*2)+1] = 1;
            int drawstart = 0;
            int drawend = 39;
            switch (secloop) {
                case 0:
                case 2:
                case 4:
                    secstart = 1;
                    map[((secloop/2) * 7)+0][0] = EFTTLWall;
                    map[((secloop/2) * 7)+1][0] = EFTVWall;
                    map[((secloop/2) * 7)+2][0] = EFTVWall;
                    map[((secloop/2) * 7)+3][0] = EFTVWall;
                    map[((secloop/2) * 7)+4][0] = EFTBLWall;
                    if(Placed[secloop+1] == 0) {
                        drawend = 40;

                        map[((secloop/2) * 7) + 0][41] = EFTTLWall;
                        map[((secloop/2) * 7) + 1][41] = EFTVWall;
                        map[((secloop/2) * 7) + 2][41] = EFTVWall;
                        map[((secloop/2) * 7) + 3][41] = EFTVWall;
                        map[((secloop/2) * 7) + 4][41] = EFTBLWall;
                    }
                    break;
                case 1:
                case 3:
                case 5:
                    secstart = 39;
                    map[((secloop/2) * 7)+0][78] = EFTTRWall;
                    map[((secloop/2) * 7)+1][78] = EFTVWall;
                    map[((secloop/2) * 7)+2][78] = EFTVWall;
                    map[((secloop/2) * 7)+3][78] = EFTVWall;
                    map[((secloop/2) * 7)+4][78] = EFTBRWall;
                    if(Placed[secloop-1] == 0) {
                        drawstart = -1;

                        map[((secloop/2) * 7) + 0][37] = EFTTRWall;
                        map[((secloop/2) * 7) + 1][37] = EFTVWall;
                        map[((secloop/2) * 7) + 2][37] = EFTVWall;
                        map[((secloop/2) * 7) + 3][37] = EFTVWall;
                        map[((secloop/2) * 7) + 4][37] = EFTBRWall;
                    }
                    break;
            }



            for(int drawsec = drawstart;drawsec < drawend; drawsec++) {
                map[((secloop/2) * 7)+0][drawsec+secstart] = EFTHWall;
                map[((secloop/2) * 7)+1][drawsec+secstart] = EFTFloor;
                map[((secloop/2) * 7)+2][drawsec+secstart] = EFTWater;
                map[((secloop/2) * 7)+3][drawsec+secstart] = EFTFloor;
                map[((secloop/2) * 7)+4][drawsec+secstart] = EFTHWall;
            }
        }
    }

////////////////////////////////////////////////////////////////////////////////

    //Now make the vertical connectors, these will be randomly chosen from
    //the existing horizontal sections.

    int vertloc[3] = {0,-1,0};
    for(int verloop = 0; verloop < 4; verloop += 2) {
        int vert[3] = {0,0,0};
        if(Placed[verloop] == 1 && Placed[verloop+2] == 1)
            vert[0] = 1;
        if(Placed[verloop] == 1 && Placed[verloop+3] == 1) vert[1] = 1;
        if(Placed[verloop+1] == 1 && Placed[verloop+2] == 1) vert[1] = 1;
        if(Placed[verloop+1] == 1 && Placed[verloop+3] == 1)
            vert[2] = 1;


        //If no horizontal sections are below this one then make a vertical
        //section in the middle
        if((Placed[0] == 1 || Placed[1] == 1) &&
        (vert[0] ==0 && vert[1] == 0 && vert[2] == 0) &&
        (Placed[4] == 1 || Placed[5] == 1)) {
            vert[1] = 1;
            if(verloop == 0) {
                vertloc[1] = 37;
                if(Placed[0] == 1 && Placed[1] == 0 && Placed[4] == 1) vertloc[1] = (rn2(33));
                if(Placed[0] == 0 && Placed[1] == 1 && Placed[5] == 1) vertloc[1] = (rn2(33)) + 42;
                if(Placed[0] == 1 && Placed[1] == 1) {
                    if(Placed[4] == 1 && Placed[5] == 1) vertloc[1] = (rn2(75));
                    if(Placed[4] == 1 && Placed[5] == 0) vertloc[1] = (rn2(33));
                    if(Placed[4] == 0 && Placed[5] == 1) vertloc[1] = (rn2(33)) + 42;
                }
            }

            if((verloop/2) == 0 || ((verloop/2) == 1 && (Placed[4] != 1 && Placed[5] != 1))) {
                map[((verloop/2)* 7)+8][vertloc[1]+0] = EFTVWall;
                map[((verloop/2)* 7)+8][vertloc[1]+1] = EFTFloor;
                //water already there
                map[((verloop/2)* 7)+8][vertloc[1]+3] = EFTFloor;
                map[((verloop/2)* 7)+8][vertloc[1]+4] = EFTVWall;

                map[((verloop/2)* 7)+9][vertloc[1]+0] = EFTVWall;
                map[((verloop/2)* 7)+9][vertloc[1]+1] = EFTFloor;
                map[((verloop/2)* 7)+9][vertloc[1]+2] = EFTWater;
                map[((verloop/2)* 7)+9][vertloc[1]+3] = EFTFloor;
                map[((verloop/2)* 7)+9][vertloc[1]+4] = EFTVWall;

                map[((verloop/2)* 7)+10][vertloc[1]+0] = EFTVWall;
                map[((verloop/2)* 7)+10][vertloc[1]+1] = EFTFloor;
                //water already there
                map[((verloop/2)* 7)+10][vertloc[1]+3] = EFTFloor;
                map[((verloop/2)* 7)+10][vertloc[1]+4] = EFTVWall;
            }
        }

        if(vert[0] == 1) vertloc[0] = rn2(33);
        if(vert[2] == 1) vertloc[2] = rn2(33) + 42;
        if(vert[1] == 1 && vertloc[1] == -1) vertloc[1] = 37;

        if(vert[0] == 1 || vert[2] == 1) vert[1] = 0;
        if(verloop == 2 && Placed[4] == 0 && Placed[5] == 0) vert[1] = 0;

        //Randomly make them (at least 1)
        int nVert = rn2(2) + 1;


        //Make the vertical section
        for(int zz=0;zz < 500;zz++) {
            int choosevert = rn2(3);
            if(vert[choosevert] == 1) {
                map[((verloop/2)* 7)+4-3][vertloc[choosevert]+2] = EFTWater;
                map[((verloop/2)* 7)+4-1][vertloc[choosevert]+2] = EFTWater;
                map[((verloop/2)* 7)+4+4][vertloc[choosevert]+2] = EFTWater;
                map[((verloop/2)* 7)+4+6][vertloc[choosevert]+2] = EFTWater;
                for(int drawloop = 0; drawloop < 4; drawloop++) {
                    map[((verloop/2)* 7)+4+drawloop][vertloc[choosevert]+0] = EFTVWall;
                    map[((verloop/2)* 7)+4+drawloop][vertloc[choosevert]+1] = EFTFloor;
                    map[((verloop/2)* 7)+4+drawloop][vertloc[choosevert]+2] = EFTWater;
                    map[((verloop/2)* 7)+4+drawloop][vertloc[choosevert]+3] = EFTFloor;
                    map[((verloop/2)* 7)+4+drawloop][vertloc[choosevert]+4] = EFTVWall;
                }
                nVert--;
            }
            if (nVert == 0) break;
        }
    }

////////////////////////////////////////////////////////////////////////////////

    //Spawn some water
    int nPuddles = rn2(3) + 2;
    for (int zz = 0; zz < 500; zz++) {
        int tempy = rn2(80);
        int tempx = rn2(80);

        if (map[tempy][tempx] == EFTWater) {
	    for(int rtemp = 0;rtemp < (rn2(15) + 20); rtemp++) {
	      int dispx = (rn2(9) - 4);
              int dispy = (rn2(5) - 2);

              if((tempx+dispx) > 79 || (tempx+dispx) < 0 ) dispx = 0;
              if((tempy+dispy) > 19 || (tempy+dispy) < 0 ) dispy = 0;
	           if(map[tempy+dispy][tempx+dispx] == EFTFloor)
	          map[tempy+dispy][tempx+dispx] = EFTWater;
	    }
            nPuddles--;
        }
        if (nPuddles == 0) break;
    }


////////////////////////////////////////////////////////////////////////////////

    //This section is for the final section of corridor to the e-room.
    //It must be done before room is made
    int finaly = 0;
    int finalx = 0;
    int finalflag = 0;

    //Find a secton of floor, which means the corridor
    //will connect to the actual sewers
    for(int zz=0; zz < 1000; zz++) {
        int tempy = rn2(20);
        int tempx = rn2(80);

        if(map[tempy][tempx] == EFTFloor) {
            finaly = tempy;
            finalx = tempx;
            finalflag = 1;
        }
        if(finalflag == 1) break;
    }
    //If it didnt randomly find a floor tile, then screw it, just hard code
    //one in
    if(finalflag != 1) {
        for(int zz=0; zz<6; zz += 2) {
            if(Placed[zz] == 1) {
                finaly = (3+(((zz-1)/4)*7));
                finalx = 1;
            } else if(Placed[(zz+1)] == 1) {
                finaly = (3+((((zz+1)-3)/4)*7));
                finalx = 77;
            }
        }
    }

////////////////////////////////////////////////////////////////////////////////

    //Make "the room", a random ellipse and place it in an empty section
    float height = 0;
    float width = 0;

    //Find an empty place
    int placex = 0, placey = 0;
    //If after 500 tries, then oh well, no room.  It cannot be forced in.
    for(int zz=0;zz<500;zz++) {
        int randtemp = rn2(6);
        if(Placed[randtemp] == 0) {
            //width and height must be done seperately.  Width is for collision
            //detection, height is done after its location is known.
            width = rn2(4) + 3;
            placey = ((randtemp/2) *7) + 2;
            placex = (rn2(20) + ((randtemp % 2) * 40)) + 4;
            int flag = 0;
            for (int collcheck=(-1*int(width));collcheck<=int(width);collcheck++)
	       if(map[placey][placex+collcheck] == EFTVWall)
	           flag = 1;
            if(flag == 0) {
                int hmod = 2;
                if(randtemp < 2) hmod = 1;
                height = rn2(hmod) + 3;
                break;
            }
        }
    }

    //Call the ellipse function to create the actual ellipse
    ellipse(width, height, placey, placex);


////////////////////////////////////////////////////////////////////////////////

    //create tunnels
    //first create the real tunnel
    int randxinit,randyinit,randx,randy,corrdist,totdist = 0;
    //corridor distance is an arbitrary length, it just states how long
    //the length of all the corridors will approximately be
    corrdist = rn2(1) + 20;

    randx = rn2(21) - 10;
    randy = rn2(13) - 6;
    //I dont want the tunnels to be running against the walls so overshooting
    //is just set to 0
    if(placex + randx > 79) randx = 0;
    if(placex + randx < 0) randx = 0;
    if(placey + randy > 19) randy = 0;
    if(placey + randy < 0) randy = 0;

    //call initial tunnel line, its spits out form the center of the ellipse
    solidline(placey, placex, randy+placey, randx+placex);

    //counter for total distance
    totdist = totdist + ((randx + randy)/2);

    randyinit = randy+placey;
    randxinit = randx+placex;

    //make the rest of the tunnels
    do {
        randx = rn2(21) - 10;
        randy = rn2(13) - 6;
        if(randxinit + randx > 79) randx = 0;
        if(randxinit + randx < 0) randx = 0;
        if(randyinit + randy > 19) randy = 0;
        if(randyinit + randy < 0) randy = 0;

        solidline(randyinit, randxinit, randyinit+randy, randxinit+randx);
        totdist = totdist + ((abs(randx + randy))/2);
        randyinit = randy + randyinit;
        randxinit = randx + randxinit;
    } while(totdist < corrdist);

    //Using the finalx and finaly variables from before, connect the
    //tunnel to the sewer
    solidline(randyinit, randxinit, finaly, finalx);
}
