/* Interhack   Multi-player Rougelike game (based on Nethack)
   Copyright (C) 1996-2000  Damian Bentley

   See license.txt for details. If you do not have a copy of the license,
   please email bendchon@mehta.anu.edu.au */

#include "object.h"
#include "log.h"
#include "color.h"

ID nextObjectId = 0;

void initObjects() {
    initAmulets();	initPotions();	initWeapons();	initFoods();
    initGems();		initScrolls();	initSpells();	initTools();
    initArmor();	initRings();	initWands();	initArtifacts();
    initGold();		initLevelObjects();
}

void loadObjects(ID i) {
    loadAmulets(i);	loadPotions(i);	loadWeapons(i);	loadFoods(i);
    loadGems(i);	loadScrolls(i);	loadSpells(i);	loadTools(i);
    loadArmor(i);	loadRings(i);	loadWands(i);	loadArtifacts(i);
    loadLevelObjects(i);
}

void saveObjects(ID i) {
    saveAmulets(i);	savePotions(i);	saveWeapons(i);	saveFoods(i);
    saveGems(i);	saveScrolls(i);	saveSpells(i);	saveTools(i);
    saveArmor(i);	saveRings(i);	saveWands(i);	saveArtifacts(i);
    saveLevelObjects(i);
}

void quitObjects(ID i) {
    quitAmulets(i);	quitPotions(i);	quitWeapons(i);	quitFoods(i);
    quitGems(i);	quitScrolls(i);	quitSpells(i);	quitTools(i);
    quitArmor(i);	quitRings(i);	quitWands(i);	quitArtifacts(i);
    quitLevelObjects(i);
}

CObject::CObject(EObjectType t, char sym, int wght, int cst) {
    objectType = t;
    id = nextObjectId++;
    weight = wght;
    symbol = sym;
    cost = cst;
    blessed = cursed = false;
    for (int x = 0; x < MAXPLAYERS; x++)
        known[x] = MAXPLAYERS;
    number = 1;
    isinlist = false;
    color = CLR_GREEN;
}

CObject::CObject(EObjectType t, char sym, int wght) {
    objectType = t;
    id = nextObjectId++;
    weight = wght;
    symbol = sym;
    cost = 0;
    blessed = cursed = false;
    for (int x = 0; x < MAXPLAYERS; x++)
        known[x] = MAXPLAYERS;
    number = 1;
    isinlist = false;
    color = CLR_GREEN;
}

CObject::CObject(EObjectType t, char sym) {
    objectType = t;
    id = nextObjectId++;
    weight = -1;
    symbol = sym;
    cost = 0;
    blessed = cursed = false;
    for (int x = 0; x < MAXPLAYERS; x++)
        known[x] = MAXPLAYERS;
    number = 1;
    isinlist = false;
    color = CLR_GREEN;
}

CObject::CObject(const CObject& rhs) {
    objectType = rhs.objectType;
    id = rhs.id;
    weight = rhs.weight;
    symbol = rhs.symbol;
    name = rhs.name;
    cost = rhs.cost;
    blessed = rhs.blessed;
    cursed = rhs.cursed;
    for (int x = 0; x < MAXPLAYERS; x++)
        known[x] = rhs.known[x];
    number = rhs.number;
    isinlist = rhs.isinlist;
    color = rhs.color;
}

// Select a random object based on array of probabilities p
// The sum of ints in p MUST be equal (or greater) to 1000 to work
int CObject::selectRandom(int p[]) {
    int r = rn2(1000);
    int x = 0;
    for (x = 0; r >= 0; x++)
        r -= p[x];
    return x-1;
}

bool CObject::BCisKnown(ID user) {
    for (int x = 0; x < MAXPLAYERS; x++)
        if (known[x] == user) return true;
    return false;
}

void CObject::BCidentify(ID user) {
    for (int x = 0; x < MAXPLAYERS; x++)
        if (known[x] == MAXPLAYERS) {
            known[x] = user;
            return;
        }
    for (int x = 1; x < MAXPLAYERS; x++)
        known[x-1] = known[x];
    known[MAXPLAYERS-1] = user;
}

bool CObject::operator== (CObject *o) {
    if (o == NULL || o->objectType != objectType)
        return false;
    switch (objectType) {
        case NullObject:log << "ERROR: invalid object\n"; exit(1);
        case Amulet:    return (CAmulet &)*this == (CAmulet *)o;
        case Armor:     return (CArmor &)*this == (CArmor *)o;
        case Artifact:  return (CArtifact &)*this == (CArtifact *)o;
        case Food:      return (CFood &)*this == (CFood *)o;
        case Gem:       return (CGem &)*this == (CGem *)o;
        case Potion:    return (CPotion &)*this == (CPotion *)o;
        case Ring:      return (CRing &)*this == (CRing *)o;
        case Scroll:    return (CScroll &)*this == (CScroll *)o;
        case Spellbook: return (CSpell &)*this == (CSpell *)o;
        case Tool:      return (CTool &)*this == (CTool *)o;
        case Wand:      return (CWand &)*this == (CWand *)o;
        case Weapon:    return (CWeapon &)*this == (CWeapon *)o;
        case Gold:      return (CGold &)*this == (CGold *)o;
        case LevelObject:return (CLevelObject &)*this == (CLevelObject *)o;
    }
    return false;
}

// This goes here because gold does not really have any location to live
// except object.h
void initGold() { };

#ifdef TEST_OBJECTS
#include "objlist.h"

int main() {
    /* Seed the random number generator and create a list  */
    seedrand();
    CObjectList list;

    /* Add a potion to the list */
    CObject * test = new CPotion();
    list.add(test);
    cout << "\nAdded potion: " << endl << list << endl;

    /* Try remove, and second remove */
    list.remove(test);
    list.remove(test);
    cout << "\nDeleted potion: " << endl << list << endl;

    /* Try adding the same object twice */
    test = new CPotion();
    list.add(test);
    list.add(test);
    cout << "\nTried double add: " << endl << list << endl;

    /* Add another object to the list */
    test = new CWeapon();
    list.add(test);
    cout << "\nAdded weapon: " << endl << list << endl;

    /* Try polymorph a list */
    list.polymorph();
    cout << "\nPolymorphed list: " << endl << list << endl;

    /* Test moving an object */
    CObjectList list2;
    cout << "\nList 2: " << endl << list2 << endl;
    moveobject(test, list, list2);
    cout << "List after move: " << endl << list << endl;
    cout << "List 2 after move: " << endl << list2 << endl;

    /* Test creating a players pack */
    CObjectList pack;
    pack.createPack('K', 0);
    cout << "\nPlayers pack: " << endl << pack << endl;

    /* Test adding two objects of exactly same type to the same list
       (should join up) */
    CObjectList join;
    join.add(new CPotion(PoSleeping));
    join.add(new CPotion(PoSleeping));
    cout << "\nJoined objects in a list: " << endl << join << endl;

    return 0;
}
#endif
