/* Interhack   Multi-player Rougelike game (based on Nethack)
   Copyright (C) 1996-2000  Damian Bentley

   See license.txt for details. If you do not have a copy of the license,
   please email bendchon@mehta.anu.edu.au */

#ifndef _object_h_
#define _object_h_

#include "system.h"
#include "string.h"
#include "location.h"
#include "color.h"

void initObjects();

/* These are for the players when they start, load, save or quit the game
   ie save what the player does and does not know about, and randomize
   descriptions for any new players after this one. */
void saveObjects(ID);
void loadObjects(ID);
void quitObjects(ID);

class CObject {
public:
    int         getWeight()     { return weight * number; };
    EObjectType getObjectType() { return objectType; };
    char        getSymbol()     { return symbol; };
    string      getName()       { return name; };
    bool        isBlessed()     { return blessed; };
    bool        isCursed()      { return cursed; };
    void        setName(string s){ name = s; };
    int         getNumber()     { return number; };
    void	setNumber(int n){ number = n; };
    void	decNumber()	{ number--; };
    int         getCost()       { return cost; };
    void        curseIt()       { blessed = false; cursed = true; };
    void        blessIt()       { blessed = true; cursed = false; };
    void        neutralIt()     { blessed = cursed = false; };
    bool	inList()	{ return isinlist; };
    void	isInList()	{ isinlist = true; };

    virtual void polymorph() = 0;
    virtual string getDescription(ID user) = 0;
    virtual int getColor(ID user) = 0;
    virtual bool canEat() = 0;
    virtual bool isKnown(ID user) = 0;
    virtual void identify(ID user) = 0;
    void BCidentify(ID user);
    virtual ~CObject() { };
    bool operator== (CObject *);

    /* Globally visible variables */
    CLocation   xy;
    ID          id;

private:
    EObjectType objectType;
    int         weight;
    char        symbol;
    string      name;
    int         cost;
    bool        blessed;
    bool        cursed;
    ID          known[MAXPLAYERS];    // for blessed / cursed knowledge
    bool	isinlist;
    int		color;

protected:
    CObject(const CObject&);
    CObject(EObjectType, char sym);
    CObject(EObjectType, char sym, int wght);
    CObject(EObjectType, char sym, int wght, int cst);
    int selectRandom(int []);
    bool BCisKnown(ID user);

    void setWeight(int w) { weight = w; };
    void setCost(int c)   { cost = c; };

    int         number;
};

class CAmulet;
class CArmor;
class CArtifact;
class CFood;
class CGem;
class CPotion;
class CRing;
class CScroll;
class CSpell;
class CTool;
class CWand;
class CWeapon;
class CLevelObject;

////////////////////////////////////////////////////////////////////////////////

enum EAmuletType {
    AmESP,AmLifeSaving,AmStrangulation,AmRestfulSleep,AmVersusPoison,
    AmChange,AmUnchanging,AmReflection,AmMagicalBreathing,
    AmImitationYendor,AmYendor};

void initAmulets();

void loadAmulets(ID);
void saveAmulets(ID);
void quitAmulets(ID);

class CAmulet : public CObject {
public:
    CAmulet();
    CAmulet(EAmuletType);
    EAmuletType getAmuletType() { return amuletType; };

    virtual void polymorph();
    virtual string getDescription(ID user);
    virtual int getColor(ID user);
    virtual bool canEat();
    virtual bool isKnown(ID user);
    virtual void identify(ID user);
    virtual bool operator== (CAmulet *);

private:
    EAmuletType amuletType;

protected:
};

////////////////////////////////////////////////////////////////////////////////

enum EPotionType {
    PoGainAbility,PoRestoreAbility,PoConfusion,PoBlindness,PoParalysis,
    PoSpeed,PoLevitation,PoHallucination,PoInvisibility,PoSeeInvisible,
    PoHealing,PoExtraHealing,PoGainLevel,PoEnlightenment,PoMonsterDetection,
    PoObjectDetection,PoGainEnergy,PoSleeping,PoFullHealing,PoPolymorph,
    PoBooze,PoSickness,PoFruitJuice,PoAcid,PoOil,
    PoWater};

void initPotions();

void loadPotions(ID);
void savePotions(ID);
void quitPotions(ID);

class CPotion : public CObject {
public:
    CPotion();
    CPotion(EPotionType);
    CPotion(EPotionType, int);
    EPotionType getPotionType() { return potionType; };
    void dip(CObject&);
    void dip(CPotion&);
    void dip(CTool&);
    void dip(CWeapon&);
    bool isDiluted() { return diluted; };
    void diluteIt() { diluted = true; };

    virtual void polymorph();
    virtual string getDescription(ID user);
    virtual int getColor(ID user);
    virtual bool canEat() { return false; };
    virtual bool isKnown(ID user);
    virtual void identify(ID user);
    virtual bool operator== (CPotion *);

private:
    void dipFinish(CObject&);

    EPotionType potionType;
    bool diluted;

protected:
};

////////////////////////////////////////////////////////////////////////////////

enum EWeaponType {
    /* missiles */
    WeArrow, WeElvenArrow, WeOrcishArrow, WeSilverArrow, WeYa, WeCrossbowBolt,
    WeDart, WeShuriken, WeBoomerang,

    /* spears */
    WeSpear, WeElvenSpear, WeOrcishSpear, WeDwarvishSpear, WeSilverSpear,
    WeJavelin, WeTrident,

    /* blades */
    WeDagger, WeElvenDagger, WeOrcishDagger, WeSilverDagger, WeAthame,
    WeScalpel, WeKnife, WeStiletto, WeWormTooth, WeCrysknife,
    WeAxe, WeBattleAxe,

    /* swords */
    WeShortSword, WeElvenShortSword, WeOrcishShortSword,
    WeDwarvishShortSword, WeScimitar, WeSilverSaber, WeBroadsword,
    WeElvenBroadsword, WeLongSword, WeTwohandedSword, WeKatana,

    /* special swords set up for artifacts */
    WeTsurugi, WeRunesword,

    /* polearms */
    /* spear-type */
    WePartisan, WeRanseur, WeSpetum, WeGlaive, WeLance,
    /* axe-type */
    WeHalberd, WeBardiche, WeVoulge, WeDwarvishMattock,
    /* curved/hooked */
    WeFauchard, WeGuisarme, WeBillGuisarme,
    /* other */
    WeLucernHammer, WeBecDeCorbin,

    /* bludgeons */
    WeMace, WeMorningStar, WeWarHammer, WeClub, WeRubberHose,
    WeQuaterstaff,
    /* two-piece */
    WeAklys, WeFlail,

    /* misc */
    WeBullwhip,

    /* bows */
    WeBow, WeElvenBow, WeOrcishBow, WeYumi, WeSling, WeCrossbow
};

void initWeapons();

void loadWeapons(ID);
void saveWeapons(ID);
void quitWeapons(ID);

class CWeapon : public CObject {
public:
    CWeapon();
    CWeapon(EWeaponType);
    CWeapon(EWeaponType, int);
    CWeapon(EWeaponType, int, int);
    EWeaponType getWeaponType() { return weaponType; };
    bool isPoisoned() { return poisoned > 0; };
    void setPoisoned(int i) { poisoned = i; };
    void unsetPoisoned() { poisoned = 0; };
    bool isDamaged() { return damage > 0; };
    void addDamage(int i) { damage += i; };

    virtual void polymorph();
    virtual string getDescription(ID user);
    virtual int getColor(ID user);
    virtual bool canEat();
    virtual bool isKnown(ID user);
    virtual void identify(ID user);
    virtual bool operator== (CWeapon *);

private:
    EWeaponType weaponType;
    int poisoned;
    int enchantment;
    int damage;

protected:
};

////////////////////////////////////////////////////////////////////////////////

enum EGemType {
    GemDilCrys, GemDiamond, GemRuby, GemSapphire, GemEmerald, GemTurquoise,
    GemAquamarine, GemAmber, GemTopaz, GemOpal, GemGarnet, GemAmethyst,
    GemJasper, GemFlourite, GemJade,

    GemBlueGlass, GemGreenGlass, GemRedGlass, GemVioletGlass, GemWhiteGlass,
    GemYellBrownGlass,

    GemLuckstone, GemFlintstone, GemLoadstone, GemRock};

void initGems();

void loadGems(ID);
void saveGems(ID);
void quitGems(ID);

class CGem : public CObject {
public:
    CGem();
    CGem(EGemType);
    CGem(EGemType, int);
    EGemType getGemType() { return gemType; };

    virtual void polymorph();
    virtual string getDescription(ID user);
    virtual int getColor(ID user);
    virtual bool canEat() { return false; };
    virtual bool isKnown(ID user);
    virtual void identify(ID user);
    virtual bool operator== (CGem *);

private:
    EGemType gemType;

protected:
};

////////////////////////////////////////////////////////////////////////////////

enum EScrollType {
    ScrMail,ScrIdentify,ScrLight,ScrBlankPaper,ScrEnchantWeapon,ScrEnchantArmor,
    ScrRemoveCurse,ScrConfuseMonster,ScrDestroyArmor,ScrFire,ScrFoodDetection,
    ScrGoldDetection,ScrMagicMapping,ScrScareMonster,ScrTeleportation,
    ScrAmnesia,ScrCreateMonster,ScrTaming,ScrCharging,ScrGenocide,
    ScrPunishment};

void initScrolls();

void loadScrolls(ID);
void saveScrolls(ID);
void quitScrolls(ID);

class CScroll : public CObject {
public:
    CScroll();
    CScroll(EScrollType);
    CScroll(EScrollType, int);
    EScrollType getScrollType() { return scrollType; };

    virtual void polymorph();
    virtual string getDescription(ID user);
    virtual int getColor(ID user);
    virtual bool canEat() { return false; };
    virtual bool isKnown(ID user);
    virtual void identify(ID user);
    virtual bool operator== (CScroll *);

private:
    EScrollType scrollType;

protected:
};

////////////////////////////////////////////////////////////////////////////////

enum ESpellType {
    SpellDig,SpellMagicMissile,SpellFireball,SpellConeOfCold,SpellSleep,
    SpellFingerOfDeath,SpellLight,SpellDetectMonsters,SpellHealing,SpellKnock,
    SpellForceBolt,SpellConfuseMonster,SpellCureBlindness,SpellDrainLife,SpellSlowMonster,
    SpellWizardLock,SpellCreateMonster,SpellDetectFood,SpellCauseFear,SpellClairvoyance,
    SpellCureSickness,SpellCharmMonster,SpellHasteSelf,SpellDetectUnseen,SpellLevitation,
    SpellExtraHealing,SpellRestoreAbility,SpellInvisibility,SpellDetectTreasure,SpellRemoveCurse,
    SpellMagicMapping,SpellIdentify,SpellTurnUndead,SpellPolymorph,SpellTeleportAway,
    SpellCreateFamiliar,SpellCancellation,SpellProtection,SpellJumping,SpellStoneToFlesh,
    SpellFlameSphere,SpellFreezeSphere,SpellBlankPaper,SpellBookOfTheDead};

void initSpells();

void loadSpells(ID);
void saveSpells(ID);
void quitSpells(ID);

class CSpell : public CObject {
public:
    CSpell();
    CSpell(ESpellType);
    ESpellType getSpellType() { return spellType; };

    virtual void polymorph();
    virtual string getDescription(ID user);
    virtual int getColor(ID user);
    virtual bool canEat() { return false; };
    virtual bool isKnown(ID user);
    virtual void identify(ID user);
    virtual bool operator== (CSpell *);

private:
    ESpellType spellType;

protected:
};

////////////////////////////////////////////////////////////////////////////////

enum EFoodType {
     FoodTripeRation,FoodCorpse,FoodEgg,FoodMeatball,FoodMeatStick,
     FoodChunkOfMeat,FoodMeatRing,

     FoodKelpFrond,FoodEucalyptusLeaf,FoodApple,FoodOrange,FoodPear,
     FoodMelon,FoodBanana,FoodCarrot,FoodWolfsbane,FoodGarlic,FoodSlimeMold,

     FoodRoyalJelly,FoodCreamPie,FoodCandyBar,FoodFortuneCookie,
     FoodPancake,FoodLambasWafer,FoodCramRation,FoodRation,
     FoodKRation,FoodCRation,FoodTin};

void initFoods();

void loadFoods(ID);
void saveFoods(ID);
void quitFoods(ID);

class CFood : public CObject { public:
    CFood();
    CFood(EFoodType);
    CFood(EFoodType, int);
    CFood(int);
    EFoodType getFoodType() { return foodType; };
    int eat();

    virtual void polymorph();
    virtual string getDescription(ID user);
    virtual int getColor(ID user);
    virtual bool canEat() { return true; };
    virtual bool isKnown(ID user);
    virtual void identify(ID user);
    virtual bool operator== (CFood *);

private:
    EFoodType foodType;
    int ammountEaten;
    int openAttempt;

protected:
};

////////////////////////////////////////////////////////////////////////////////

enum ERingType {
    RingAdornment,RingGainStrength,RingGainConstitution,RingIncreaseAccuracy,RingIncreaseDamage,
    RingProtection,RingRegeneration,RingSearching,RingStealth,RingSustainAbility,
    RingLevitation,RingHunger,RingAggravateMonster,RingConflict,RingWarning,
    RingPoisonResistance,RingFireResistance,RingColdResistance,RingShockResistance,RingFreeAction,
    RingSlowDigestion,RingTeleportation,RingTeleportControl,RingPolymorph,RingPolymorphControl,
    RingInvisibility,RingSeeInvisible,RingProtFromShapeChangers};

void initRings();

void loadRings(ID);
void saveRings(ID);
void quitRings(ID);

class CRing : public CObject {
public:
    CRing();
    CRing(ERingType);
    ERingType getRingType() { return ringType; };
    int getCharge() {
        if (ringType == RingAdornment || ringType == RingProtection ||
            ringType == RingGainStrength || ringType == RingIncreaseDamage)
            return charge;
        return 0; };

    virtual void polymorph();
    virtual string getDescription(ID user);
    virtual int getColor(ID user);
    virtual bool canEat();
    virtual bool isKnown(ID user);
    virtual void identify(ID user);
    virtual bool operator== (CRing *);

private:
    ERingType ringType;
    int charge;

protected:
};

////////////////////////////////////////////////////////////////////////////////

enum EArmorType {
    /* Helmets */
    HelmElvenLeather, HelmOrcish, HelmDwarvishIron, HelmFedora,
    HelmCornutham, HelmDunceCap, HelmDentedPot, HelmHelmet,
    HelmBrilliance, HelmOppositeAlignment, HelmTelepathy,
    /* Dragon suites */
    DragonGray, DragonSilver, DragonShimmering, DragonRed, DragonWhite,
    DragonOrange, DragonBlack, DragonBlue, DragonGreen, DragonYellow,
    DragonScaleGray, DragonScaleSilver, DragonScaleShimmering, DragonScaleRed,
    DragonScaleWhite, DragonScaleOrange, DragonScaleBlack, DragonScaleBlue,
    DragonScaleGreen, DragonScaleYellow,
    /* Suits */
    ArmPlate, ArmCrystalPlate, ArmBronzePlate, ArmSplint, ArmBanded,
    ArmDwarvishMithrilCoat, ArmElvenMithrilCoat, ArmChain, ArmOrcishChain,
    ArmScale, ArmStudedLeather, ArmRing, ArmOrcishRing, ArmLeatherArmor,
    ArmLeatherJacket,
    /* Shirts */
    ShirtHawaiian, ShirtT,
    /* Cloaks */
    CloakMummyWrapping, CloakElven, CloakOrchish, CloakDwarvish,
    CloakOilskin, CloakRobe, CloakSmock, CloakProtection,
    CloakInvisibility, CloakMagicResistance, CloakDisplacement,
    /* Shields */
    ShieldSmall, ShieldElven, ShieldUrukHai, ShielOrcish,
    ShieldLarge, ShieldDwarvishRound, ShieldReflection,
    /* Gloves */
    GloveLeather, GloveFumbling, GlovePower, GloveDexterity,
    /* Boots */
    BootsLow, BootsIron, BootsHigh, BootsSpeed, BootsWaterWalking,
    BootsJumping, BootsElven, BootsKicking, BootsFumble, BootsLevitation
};

void initArmor();

void loadArmor(ID);
void saveArmor(ID);
void quitArmor(ID);

class CArmor : public CObject {
public:
    CArmor();
    CArmor(EArmorType);
    CArmor(EArmorType, int, int);
    EArmorType getArmorType() { return armorType; };
    int getAC();

    virtual void polymorph();
    virtual string getDescription(ID user);
    virtual int getColor(ID user);
    virtual bool canEat();
    virtual bool isKnown(ID user);
    virtual void identify(ID user);
    virtual bool operator== (CArmor *);

private:
    EArmorType armorType;
    int enchantment;

protected:
};

////////////////////////////////////////////////////////////////////////////////

enum EWandType {
    WandLight, WandNothing, WandDigging, WandLocking, WandMagicMissile,
    WandMakeInvis, WandOpening, WandProbing, WandDoorDetection,
    WandSlowMonster, WandSpeedMonster, WandStriking, WandUndeadTurning,
    WandCold, WandFire, WandLightning, WandSleep, WandCancellation,
    WandCreateMonster, WandPolymorph, WandTeleportation, WandDeath,
    WandWishing};

void initWands();

void loadWands(ID);
void saveWands(ID);
void quitWands(ID);

class CWand : public CObject {
public:
    CWand();
    CWand(EWandType);
    EWandType getWandType() { return wandType; };
    int getCharge() { return charge; };

    virtual void polymorph();
    virtual string getDescription(ID user);
    virtual int getColor(ID user);
    virtual bool canEat();
    virtual bool isKnown(ID user);
    virtual void identify(ID user);
    virtual bool operator== (CWand *);

private:
    EWandType wandType;
    int charge;

protected:
};

////////////////////////////////////////////////////////////////////////////////
enum EToolType {
    // Containers
    Sack, LargeBox, Chest, IceBox, BagOfHolding, BagOfTricks, OilskinSack,
    // Unlocking tools
    CreditCard, LockPick, SkeletonKey,
    // LightSources
    TallowCandle, WaxCandle, BrassLantern, OilLamp, MagicLamp,
    // Instruments
    TinWhistle, MagicWhistle, Bugle, WoodenFlute, MagicFlute, TooledHorn,
    FrostHorn, FireHorn, HornOfPlenty, LeatherDrum, DrumOfEarthquake,
    WoodenHarp, MagicHarp, Bell,
    // Traps
    Beartrap, LandMine,
    // WeaponTools
    PickAxe, UnicornHorn,
    // OtherTools
    ExpensiveCamera, Mirror, CrystalBall, Blindfold, Towel,
    Stethoscope, TinningKit, TinOpener, CanOfGrease, Figurine,
    MagicMarker,
    // Tools for animals
    Leash, Bridle, Saddle, SaddleBag, SaddleBagOfHolding,
    SaddleBagOfTricks
};

void initTools();

void loadTools(ID);
void saveTools(ID);
void quitTools(ID);

class CObjectList;

class CTool : public CObject {
public:
    CTool();
    CTool(EToolType);
    CTool(EToolType, int, int);
    EToolType getToolType() { return toolType; };
    CObjectList * getContents() { return contents; };

    virtual void polymorph();
    virtual string getDescription(ID user);
    virtual int getColor(ID user);
    virtual bool canEat();
    virtual bool isKnown(ID user);
    virtual void identify(ID user);
    virtual bool operator== (CTool *);

private:
    EToolType toolType;
    CObjectList * contents;

protected:
};

////////////////////////////////////////////////////////////////////////////////

enum EArtifactType {
    Excalibur, Stormbringer, Mjollnir, Cleaver, Grimtooth,
    Orcrist, Sting, Magicbane, Frostbrand, Firebrand,
    Dragonbane, Demonbane, Werebane, Grayswandir, Giantslayer,
    Ogresmasher, Trollsbane, Vorpalblade, SnickerSnee, Sunsword,
    OrbOfDetection, HeartOfAhriman, SceptreOfMight, StaffOfAesculapius,
    MagicMirrorOfMerlin,
    EyesOfTheOverworld, MitreOfHoliness, LongbowOfArtemis,
    MasterKeyOfThievery, TsurugiOfMuramasa,
    YendorianExpressCard, OrbOfFate, EyeOfTheAethiopica,
    
    BellOpening, Candelabrum, BookOfDead, AmuletOfYendor
};

void initArtifacts();

void loadArtifacts(ID);
void saveArtifacts(ID);
void quitArtifacts(ID);

class CArtifact : public CObject {
public:
    // Can only create a specific type of artifact
    CArtifact(EArtifactType);
    EArtifactType getArtifactType() { return artifactType; };
    CObject * getArtifact() { return artifact; };

    virtual void polymorph();
    virtual string getDescription(ID user);
    virtual int getColor(ID user);
    virtual bool canEat();
    virtual bool isKnown(ID user);
    virtual void identify(ID user);
    virtual bool operator== (CArtifact *);

private:
    EArtifactType artifactType;
    CObject * artifact;

protected:
};

////////////////////////////////////////////////////////////////////////////////

void initGold();

class CGold : public CObject {
public:
    CGold() : CObject(Gold, '$') { gold = 100; polymorph(); };
    CGold(int g) : CObject(Gold, '$', g) { gold = g; };
    int getGold() { return gold; };

    virtual void polymorph() { gold = rn2(gold); setWeight(gold); };
    virtual string getDescription(ID user) { user = 0; string s; s = "gold"; return s; };
    virtual int getColor(ID user) { return CLR_YELLOW; };
    virtual bool canEat() { return true; };
    virtual bool isKnown(ID user) { user = 0; return true; };
    virtual void identify(ID user) { user = 0; };
    virtual bool operator== (CGold *) { return true; };

private:
    int gold;

protected:
};

////////////////////////////////////////////////////////////////////////////////

enum ELevelObjectType {
    ELOTTrap, ELOTFountain, ELOTDoor, ELOTSink, ELOTGrave, ELOTThrone,
    ELOTAltar};

void initLevelObjects();

void loadLevelObjects(ID);
void saveLevelObjects(ID);
void quitLevelObjects(ID);

class CLevelObject : public CObject {
public:
    CLevelObject(ELevelObjectType);

    virtual void polymorph();
    virtual string getDescription(ID);
    virtual int getColor(ID);
    virtual bool canEat() { return false; };
    virtual bool isKnown(ID);
    virtual void identify(ID);
    virtual bool operator== (CLevelObject *);

private:
    ELevelObjectType levelobjType;
protected:
};

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

#endif
