/* Interhack   Multi-player Rougelike game (based on Nethack)
   Copyright (C) 1996-2000  Damian Bentley

   See license.txt for details. If you do not have a copy of the license,
   please email bendchon@mehta.anu.edu.au */

#include "plist.h"
#include "log.h"

CPlayerList players;

CPlayer * plist[MAXPLAYERS];

static int exclude = -1;

void initPlayers() {
    int i;
    for (i = 0; i < MAXPLAYERS; i++)
        plist[i] = NULL;
}

CSock& operator<< (CSock& s, CPlayerList l) {
    int count = 0;
    for (int i = 0; i < MAXPLAYERS; i++)
        if (i != exclude && plist[i] != NULL) count++;
    s << count;
    for (int i = 0; i < MAXPLAYERS; i++)
        if (i != exclude && plist[i] != NULL) {
            CPlayer * p = plist[i];
            s << p->id << p->getName() << (int)p->attr.ptype
              << p->getDungeon() << p->getLevelNum() << p->getFlags()
              << p->xy << p->getSymbol();
        }
    return s;
}

void CPlayerList::user_list() {
    int i, count = 0;
    log << "-------------------------------------------------------------\n";
    log << "name-type(id)@location DLvl HP(Max) Exp T:STimer:forced Flags\n";
    count = 0;
    for (i = 0; i < MAXPLAYERS; i++)
        if (plist[i] != NULL) {
            CPlayer * p = plist[i];
            log << p->getName() << '-' << p->attr.ptype
                 << '(' << p->id << ")@"
                 << p->getFrom() << " T:" << p->STTimer << ':' << p->STForcedRemain;
            if (p->isInvisible()) log << " invis";
            if (p->isAggrivate()) log << " agrivate";
            if (p->isBlind()) log << " blind";
            if (p->isHallucinating()) log << " hallu";
            if (p->canSeeInvisible()) log << " see_invis";
            if (p->isConfused()) log << " conf";
            if (p->isLevitating()) log << " levi";
            if (p->canDetectMonsters()) log << " detect_mon";
            if (p->canDetectObjects()) log << " detect_obj";
            if (p->isParalised()) log << " paralysed";
            if (p->isSleeping()) log << " sleeping";
            if (p->isFast()) log << " fast";
            if (p->isWizard()) log << " wiz";
            log << '\n';
            count++;
        }
    if (count >= 0) log << "(Total number of players: " << count << ")\n";
    log << "-------------------------------------------------------------\n";
}

/* Check to see if name n is already conntected
   return 0 if they are not on, 1 if they are logged on */
int CPlayerList::already_on(string n) {
    for (int i = 0; i < MAXPLAYERS; i++)
        if (plist[i] != NULL)
            if (strcmp(plist[i]->getName(), n.getStr()) == 0) return 1;
    return 0;
}

/* Send communication to all players on same level */
void CPlayerList::talkMsg(string s, int id) {
    for (int i = 0; i < MAXPLAYERS; i++)
        if (plist[i] != NULL &&
            plist[id]->dlevel == plist[i]->dlevel)
            *(plist[i]->sock) << p_talk << s << p_end;
}

/* Send communication to all players in dungeon */
void CPlayerList::yellMsg(string s) {
    for (int i = 0; i < MAXPLAYERS; i++)
        if (plist[i] != NULL)
            *(plist[i]->sock) << p_yell << s << p_end;
}

/* Send communication to a specific player */
void CPlayerList::chatMsg(string s, int id) {
    if (plist[id] == NULL)
        log << "ERROR: chatMsg to unknown player\n";
    else
        *(plist[id]->sock) << p_chat << s << p_end;
}

/* Send communication to players on a level...
   pmsg gets sent to the player, omsgh gets sent to
   hallucinating players that can see id, omsg is
   sent to non-hallucinating players. */
void CPlayerList::seenMsg(string omsg, string omsgh, string pmsg, int id) {
    if (plist[id] == NULL) {
        log << "ERROR: seenMsg to unknown player\n";
        return;
    }
    CLevel * mylev = plist[id]->dlevel;
    CLocation myloc = plist[id]->xy;
    for (int i = 0; i < MAXPLAYERS; i++)
        /* First find players on same level */
        if (plist[i] != NULL && plist[i]->dlevel == mylev) {
            /* If this is the player, send their part of the msg */
            if (i == id)
                *(plist[id]->sock) << p_topline << pmsg << p_end;
            else {
                CLocation l = plist[i]->xy;
                if (!(mylev->blocked_path(myloc.x, myloc.y,
                    l.x, l.y)) && !plist[i]->isBlind()) {
                    if (plist[i]->isHallucinating())
                        *(plist[i]->sock) << p_topline << omsgh << p_end;
                    else
                        *(plist[i]->sock) << p_topline << omsg << p_end;
                }
            }
        }
}

/* Send communication to players on a level... (drop and pickup)
   o is the object, id is the player */
void CPlayerList::dpMsg(string action, CObject *o, int id) {
    if (plist[id] == NULL) {
        log << "ERROR: seenMsg to unknown player\n";
        return;
    }
    CLevel * mylev = plist[id]->dlevel;
    CLocation myloc = plist[id]->xy;
    for (int i = 0; i < MAXPLAYERS; i++)
        /* First find players on same level */
        if (plist[i] != NULL && plist[i]->dlevel == mylev && i != id) {
            CLocation l = plist[i]->xy;
            if (!(mylev->blocked_path(myloc.x, myloc.y,
                l.x, l.y)) && !plist[i]->isBlind()) {
                string s;
                s = plist[id]->getName(); s += ' ';
                s += action; s += ' ';
                if (plist[i]->isHallucinating()) {
                    s += "HALLU OBJ.";
                    *(plist[i]->sock) << p_topline << s << p_end;
                } else {
                    s += o->getDescription(i); s += '.';
                    *(plist[i]->sock) << p_topline << s << p_end;
                }
            }
        }
}

/* The the movement of player id to other players on same level */
void CPlayerList::sendMovement(int id) {
    if (plist[id] == NULL)
        log << "ERROR: sendMovement to unknown player\n";
    else {
        CLocation xy = plist[id]->xy;
        CLevel *l = plist[id]->dlevel;
        for (int i = 0; i < MAXPLAYERS; i++)
            if (plist[i] != NULL /*&& i != id*/ && plist[i]->dlevel == l)
                *(plist[i]->sock) << p_move << id << xy << p_end;
    }
}

/* Returns true if any players on same level as id is on location l */
bool CPlayerList::sameLocation(int id, CLocation loc) {
    CLevel * plvl = plist[id]->dlevel;
    for (int i = 0; i < MAXPLAYERS; i++)
        if (plist[i] != NULL && i != id &&
            plist[i]->dlevel == plvl &&
            plist[i]->xy == loc)
            return true;
    return false;
}

void CPlayerList::sendToAll() {
    sendToAllExclude(-1);
}

void CPlayerList::sendToAllExclude(int pn) {
    exclude = pn;
    for (int i = 0; i < MAXPLAYERS; i++)
        if (i != pn && plist[i] != NULL)
            *(plist[i]->sock) << p_list << players << p_end;
    for (int i = 0; i < MAXPLAYERS; i++)
        if (i != pn && plist[i] != NULL)
            sendMovement(i);
    exclude = -1;
}

/* Send player id flags to all players on the level */
void CPlayerList::sendFlags(int id) {
    log << "TODO: send players flags\n";
}

/* Player p hit a monster m */
void CPlayerList::phitm(CPlayer *p, MonsterType *m) {
    /*CLevel * mylev = p->dlevel;
    CLocation myloc = p->xy;
    //string str1 = p->getName();
    if (m->mhp < 1) str1 += " kills the ";
    else            str1 += " hits the ";
    str1 += m->SName;
    for (int i = 0; i < MAXPLAYERS; i++)
        if (plist[i] != NULL && plist[i]->dlevel == p->dlevel) {
            if (plist[i] == p) {
                string str2;
                if (m->mhp < 1) str2 = "You kill the ";
                else            str2 = "You hit the ";
                str2 += m->SName;
                *(p->sock) << p_hit << str2 << m->IID
                           << (char)m->isAlive() << p_end;
            } else {
                CLocation l = plist[i]->xy;
                if (!(mylev->blocked_path(myloc.x, myloc.y,
                    l.x, l.y)) && !plist[i]->isBlind()) {
                    *(plist[i]->sock) << p_hit << str1 << m->IID
                                      << (char)m->isAlive() << p_end;
                }
            }
        }   */
}

/* Returns the closest player to location loc on dungeon level lev */
CPlayer * CPlayerList::assignClosest(CLevel *lev, CLocation loc) {
    CPlayer *temp = NULL;
    int tempdist = 1000;
    for (int i = 0; i < MAXPLAYERS; i++)
        if (plist[i] != NULL && plist[i]->dlevel == lev) {
            if (temp == NULL) {
                temp = plist[i];
                tempdist = plist[i]->xy % loc;
                continue;
            }
            if (plist[i]->xy % loc < tempdist) {
                temp = plist[i];
                tempdist = plist[i]->xy % loc;
                if (tempdist < 2) break;
            }
        }
    return temp;
}

/* Update the monster mtmp for all players on level lev */
void CPlayerList::updateMonster(CLevel *lev, MonsterType *mtmp) {
    for (int i = 0; i < MAXPLAYERS; i++)
        if (plist[i] != NULL && plist[i]->dlevel == lev)
            *(plist[i]->sock) << p_monster << mtmp->IID
                              << mtmp->xy << p_end;
}

/* Player p dropped object o */
void CPlayerList::drop(CPlayer *p, CObject *o) {
    CLevel * lev = p->dlevel;
    for (int i = 0; i < MAXPLAYERS; i++)
        if (plist[i] != NULL && plist[i]->dlevel == lev &&
            plist[i] != p)
            *(plist[i]->sock) << p_drop << o->getDescription(plist[i]->id)
                << o->id << o->getWeight()
                << (char)o->getObjectType() << o->xy
                << o->getColor(plist[i]->id) << p_end;
    dpMsg("drops", o, p->id);
}

/* Player p picked up object o
   Only need to know which object was picked up, no details needed. */
void CPlayerList::pickup(CPlayer *p, CObject *o) {
    CLevel * lev = p->dlevel;
    for (int i = 0; i < MAXPLAYERS; i++)
        if (plist[i] != NULL && plist[i]->dlevel == lev &&
            plist[i] != p)
            *(plist[i]->sock) << p_pickup << o->id << p_end;
    dpMsg("picks up", o, p->id);
}

/******************************************************************************/
/* Sureal time stuff */

Timeval LEADWAY = THELEADWAY;

/* sleepTime returns the minimum sleep time until the next expected action
   it returns 5.0 secs when no action is expected (this is used in
   tcpip.cc to check for dead connections) */
Timeval st;
Timeval * CPlayerList::sleepTime() {
    Timeval zt; ih_timerclear(&zt);
    Timeval ct; gettimeofday(&ct, NULL);
    int stSet = 0;
    ih_timerclear(&st);

    for (register int i = 0; i < MAXPLAYERS; i++) {
        if (plist[i] != NULL && plist[i]->STTimer >= 0) {
            if (ih_timercmp(plist[i]->STNext, ct, >) &&
                (ih_timercmp(plist[i]->STNext, st, <) || stSet == 0)) {
                st = plist[i]->STNext;
                stSet = 1;
            }
            if (ih_timercmp(plist[i]->STForced, ct, >) &&
                (ih_timercmp(plist[i]->STForced, st, <) || stSet == 0)) {
                st = plist[i]->STForced;
                stSet = 1;
            }
        }
    }

    ih_timersub(&st, ct);
    if (ih_timercmp(st, zt, <)) {
        st.tv_sec = 5;
        st.tv_usec = 0;
    }
#ifdef DEBUG_SUREAL
    log << "STDEBUG: Sleeping for " << ih_timershow(st) << '\n';
#endif
    return &st;
}

/* Player i has moved, returns true if can move
   this function is only called when the actual player has moved */
bool CPlayerList::doMove(int i) {
    Timeval ct; gettimeofday(&ct, NULL);
    Timeval zt; ih_timerclear(&zt);

    if (plist[i] == NULL) return false;

    /* If player i is not on a timer, all moves are valid */
    if (plist[i]->STTimer < 0) return true;

    /* Check we are not moving too fast in sureal time */
    if (ih_timercmp(plist[i]->STNext, ct, >)) {
        plist[i]->backlogged++;
        return false;
    }
    if (plist[i]->backlogged > 0)
        plist[i]->backlogged--;

    /* Setup the next valid move time */
    ih_timerclear(&(plist[i]->STNext));
    ih_timeradd(&(plist[i]->STNext), ct);
    ih_timeradd(&(plist[i]->STNext), LEADWAY);
    /* TODO: base timeout on player speeds too! */
    ih_timeradd(&(plist[i]->STNext), plist[plist[i]->STTimer]->STTimeout);

    /* Because player has moved, ignore one forced move
       and reset the next forced move time */
    bool clearedForced = false;
    if (plist[i]->STForcedRemain > 0) {
        plist[i]->STForcedRemain--;
        ih_timerclear(&(plist[i]->STForced));
        if (plist[i]->STForcedRemain > 0) {
            ih_timeradd(&(plist[i]->STForced), ct);
            ih_timeradd(&(plist[i]->STForced), LEADWAY);
            /* TODO: base timeout on player speeds too! */
            ih_timeradd(&(plist[i]->STForced), plist[i]->STTimeout);
        } else
            clearedForced = true;
    }

    /* If other players are on the same timer, force them to move */
    if (!clearedForced) {
    int mytimer = plist[i]->STTimer;
    for (int j = 0; j < MAXPLAYERS; j++)
        if (plist[j] != NULL && i != j && plist[j]->STTimer == mytimer) {
            plist[j]->STForcedRemain++;
            ih_timerclear(&(plist[j]->STForced));
            ih_timeradd(&(plist[j]->STForced), ct);
            ih_timeradd(&(plist[j]->STForced), LEADWAY);
            /* TODO: base timeout on player speeds too! */
            ih_timeradd(&(plist[j]->STForced), plist[j]->STTimeout);
        }
    }

    return true;
}

/* Force a player to move */
void CPlayerList::forcedMoves() {
    Timeval ct; gettimeofday(&ct, NULL);
    Timeval zt; ih_timerclear(&zt);

    /* If a player has been forced to move, move them, and adjust things */
    for (int i = 0; i < MAXPLAYERS; i++)
        if (plist[i] != NULL && plist[i] >= 0 &&
            ih_timercmp(plist[i]->STForced, ct, <) &&
            ih_timercmp(plist[i]->STForced, zt, !=)) {
#ifdef DEBUG_SUREAL
            log << "STDEBUG: Player " << i << " forced to move\n";
#endif
            plist[i]->STForcedRemain--;
            ih_timerclear(&(plist[i]->STForced));
            if (plist[i]->STForcedRemain > 0) {
                ih_timeradd(&(plist[i]->STForced), ct);
                ih_timeradd(&(plist[i]->STForced), LEADWAY);
                ih_timeradd(&(plist[i]->STForced), plist[i]->STTimeout);
            } else
                plist[i]->STForcedRemain = 0;

            /* Make the forced move, set the time for the next valid move */
            plist[i]->forcedMove();
            if (plist[i]->STTimer >= 0) {
                ih_timerclear(&(plist[i]->STNext));
                ih_timeradd(&(plist[i]->STNext), ct);
                ih_timeradd(&(plist[i]->STNext), LEADWAY);
                ih_timeradd(&(plist[i]->STNext), plist[plist[i]->STTimer]->STTimeout);
            }
        }
}

/* Make any moves player has banked up */
void CPlayerList::makeMoves() {
    Timeval ct; gettimeofday(&ct, NULL);
    Timeval zt; ih_timerclear(&zt);

    for (int i = 0; i < MAXPLAYERS; i++)
        if (plist[i] != NULL && plist[i]->STTimer >= 0 &&
            ih_timercmp(plist[i]->STNext, ct, <) &&
            ih_timercmp(plist[i]->STNext, zt, !=) &&
            ih_timercmp(plist[i]->STForced, zt, ==) &&
            plist[i]->backlogged > 0) {
#ifdef DEBUG_SUREAL
            log << "STDEBUG: Player " << i << " clearing backlog\n";
#endif
            plist[i]->incomming_data_backlog();
    }
}

/* Go through and update the timers whenever a player logs in / out */
void CPlayerList::arrangeTimers() {
    /* Go through each player. Players on different levels, or more than
       16 units away from other players are on their own timers */
    for (int i = 0; i < MAXPLAYERS; i++)
        if (plist[i] != NULL) {
            plist[i]->STTimer = -1;
            for (int j = 0; j < MAXPLAYERS; j++)
                if (plist[j] != NULL && j != i &&
                    plist[i]->dlevel == plist[j]->dlevel) {
                    if (plist[i]->xy % plist[j]->xy <= 16) {
                        plist[i]->STTimer = plist[j]->STTimer = i;
                        break;
                    }
                }
        }
}

/* Go through and change timers if a player moved */
void CPlayerList::arrangeTimersM() {
    /* Go through each player. Players on different levels, or more than
       16 units away from other players are on their own timers */
    for (int i = 0; i < MAXPLAYERS; i++)
        if (plist[i] != NULL) {
            plist[i]->STTimer = -1;
            for (int j = 0; j < MAXPLAYERS; j++)
                if (plist[j] != NULL && j != i &&
                    plist[i]->dlevel == plist[j]->dlevel) {   
                    if (plist[i]->xy % plist[j]->xy <= 16) {
                        plist[i]->STTimer = plist[j]->STTimer = i;
                        break;
                    }
                }
        }
}
