/* Interhack   Multi-player Rougelike game (based on Nethack)
   Copyright (C) 1996-2000  Damian Bentley

   See license.txt for details. If you do not have a copy of the license,
   please email bendchon@mehta.anu.edu.au */

/* Finally a decent data send / receive system. With this implimentation,
   all data sending / receving is done here. tcpip is used for the server
   dealing with multiple connection. To send a message, you MUST start it
   with the protocol, and end it with p_end. Eg
       *sock << p_type << 'C' << p_end;        */

#include <unistd.h>
#ifdef SERIAL
#include <termios.h>
#endif /* SERIAL */

#include "sock.h"
#include "log.h"

/* Create a socket. Because the socket has already been open, we store the
   socket number, and set the buffers up */
CSock::CSock(int fd) {
    sockfd = fd;
    sending = FALSE;
    outbuf = outptr = NULL;
    inbuf = inptr = NULL;
    if (outbuf == NULL) {
        outbuf = new char[BUFFER_SIZE];
        outptr = outbuf+2;
    }
    if (inbuf == NULL) {
        inbuf = new char[BUFFER_SIZE];
        inptr = inbuf;
        inread = inbuf + 2;
    }
    if (outbuf == NULL || inbuf == NULL) {
        log << "ERROR: Sock::Sock- malloc for sock buffers\n";
        exit(1);
    }
}

/* Delete a socket and free up its buffer. */
CSock::~CSock() {
    if (outbuf != NULL)
        delete [] outbuf;
    outbuf = outptr = NULL;
    if (inbuf != NULL)
        delete [] inbuf;
    inbuf = inptr = NULL;
}

/* Deal with control signals */
CSock& operator<< (CSock& sock, enum cont cr) {
    /* Send the message */
    if (cr == p_end) {
        if (sock.sending == FALSE) {
            log << "ERROR: op<<(Sock&, enum cont): Message already sent\n";
            exit(1);
        }
        sock.sending = FALSE;
        sock.send_msg();
    } else
        log << "ERROR: op<<(Sock&, enum cont: Unknown socket control\n";
    return sock;
}

/* Send a string, \0 terminated */
CSock& operator<< (CSock& sock, string s) {
    if (sock.sending == FALSE) {
        log << "ERROR: op<< (CSock&, int) Use *sock << [protocol] first\n";
        exit(1);
    }
    memcpy(sock.outptr, s.getStr(), s.size());
    sock.outptr += s.size();
    if (sock.outbuf + BUFFER_SIZE <= sock.outptr) {
        log << "WARNING: buffer overflow op<< (CSock&, string)\n";
        exit(1);
        sock.outptr = sock.outbuf + BUFFER_SIZE - 1;
    }
    *(sock.outptr) = '\0';
    sock.outptr++;
    return sock;
}

/* Receive a string, \0 terminated */
CSock& operator>> (CSock& sock, string & s) {
    if (sock.inbuf + BUFFER_SIZE <= sock.inread) {
        log << "WARNING: buffer overflow op>> (CSock&, string)\n";
        exit(1);
        sock.inread = sock.inbuf + BUFFER_SIZE - 1;
    }
    if (mystrlen(sock.inread) < 1)
        return sock;
    s = sock.inread;
    sock.inread += (s.size()+1);
    return sock;
}

/* Send a single character */
CSock& operator<< (CSock& sock, char c) {
    if (sock.sending == FALSE) {
        log << "ERROR: op<< (CSock&, char) Use *sock << [protocol] first\n";
        exit(1);
    }
    if (sock.outbuf + BUFFER_SIZE <= sock.outptr) {
        log << "WARNING: buffer overflow op<< (CSock&, char)\n";
        exit(1);
        sock.outptr = sock.outbuf + BUFFER_SIZE - 1;
    }
    *(sock.outptr) = c;
    sock.outptr++;
    return sock;
}

/* Receive a single character */
CSock& operator>> (CSock& sock, char &c) {
    if (sock.inbuf + BUFFER_SIZE <= sock.inread) {
        log << "WARNING: buffer overflow op>> (CSock&, char)\n";
        exit(1);
        sock.inread = sock.inbuf + BUFFER_SIZE - 1;
    }
    c = *(sock.inread);
    sock.inread++;
    return sock;
}

/* Send a boolean */
/*CSock& operator<< (CSock& sock, bool b) {
    if (sock.sending == FALSE) {
        log << "ERROR: op<< (CSock&, bool) Use *sock << [protocol] first\n";
        exit(1);
    }
    if (sock.outbuf + BUFFER_SIZE <= sock.outptr) {
        log << "WARNING: buffer overflow op<< (CSock&, char)\n";
        exit(1);
        sock.outptr = sock.outbuf + BUFFER_SIZE - 1;
    }
    *(sock.outptr) = (char)b;
    sock.outptr++;
    return sock;
}*/

/* Receive a boolean */
/*CSock& operator>> (CSock& sock, bool &b) {
    if (sock.inbuf + BUFFER_SIZE <= sock.inread) {
        log << "WARNING: buffer overflow op>> (CSock&, bool)\n";
        exit(1);
        sock.inread = sock.inbuf + BUFFER_SIZE - 1;
    }
    b = (bool)(*(sock.inread));
    sock.inread++;
    return sock;
}*/

/* Send an integer. We can not use the << and >> shift operators because
   they are machine dependant. Some machines will insert 0 when shifting,
   and other machines insert the sign flag. I know this is messy, I will
   change it some day when I can be bothered thinking. */
CSock& operator<< (CSock& sock, int i) {
    int signedbit = 0;

    if (sock.sending == FALSE) {
        log << "ERROR: op<< (CSock&, int) Use *sock << [protocol] first\n";
        exit(1);
    }
    if (i < 0) {
        signedbit = 1;
        i = -i;
    }
    if (sock.outbuf + BUFFER_SIZE <= sock.outptr + 5) {
        log << "WARNING: buffer overflow op<< (CSock&, int)\n";
        exit(1);
        sock.outptr = sock.outbuf + BUFFER_SIZE - 5;
    }
    *(sock.outptr++) = (i / (256 * 256 * 256)) - 128;
    *(sock.outptr++) = (i / (256 * 256)) - 128;
    *(sock.outptr++) = (i / 256) - 128;
    *(sock.outptr++) = (i % 256) - 128;
    *(sock.outptr++) = signedbit;

    return sock;
}

CSock& operator<< (CSock& sock, ID i) {
    if (sock.sending == FALSE) {
        log << "ERROR: op<< (CSock&, int) Use *sock << [protocol] first\n";
        exit(1);
    }
    if (sock.outbuf + BUFFER_SIZE <= sock.outptr + 4) {
        log << "WARNING: buffer overflow op<< (CSock&, int)\n";
        exit(1);
        sock.outptr = sock.outbuf + BUFFER_SIZE - 4;
    }
    *(sock.outptr++) = (i / (256 * 256 * 256)) - 128;
    *(sock.outptr++) = (i / (256 * 256)) - 128;
    *(sock.outptr++) = (i / 256) - 128;
    *(sock.outptr++) = (i % 256) - 128;

    return sock;
}

/* Receive an integer */
CSock& operator>> (CSock& sock, int &i) {
    int signedbit = 0;

    if (sock.inbuf + BUFFER_SIZE <= sock.inread) {
        log << "WARNING: buffer overflow op>> (CSock&, int)\n";
        exit(1);
        sock.inread = sock.inbuf + BUFFER_SIZE - 5;
    }
    i = 0;
    i += (*(sock.inread++) + 128) * (256 * 256 * 256);
    i += (*(sock.inread++) + 128) * (256 * 256);
    i += (*(sock.inread++) + 128) * 256;
    i += (*(sock.inread++) + 128);
    if ((signedbit = *(sock.inread++)) == 1)
        i = -i;
    return sock;
}

CSock& operator>> (CSock& sock, ID &i) {
    if (sock.inbuf + BUFFER_SIZE <= sock.inread) {
        log << "WARNING: buffer overflow op>> (CSock&, int)\n";
        exit(1);
        sock.inread = sock.inbuf + BUFFER_SIZE - 4;
    }
    i = 0;
    i += (*(sock.inread++) + 128) * (256 * 256 * 256);
    i += (*(sock.inread++) + 128) * (256 * 256);
    i += (*(sock.inread++) + 128) * 256;
    i += (*(sock.inread++) + 128);
    return sock;
}

/* Send the protocol number */
CSock& operator<< (CSock& sock, Protocol pnum) {
    /* Set up the buffer for a new message. */
    if (sock.sending == TRUE) {
        log << "ERROR: op<<(Sock&, Protocol: Already sending message\n";
        exit(1);
    }
    sock.sending = TRUE;
    sock.outptr = sock.outbuf + 2;

    /* Send the protocol */
    *(sock.outptr) = (char)pnum;
    sock.outptr++;
    return sock;
}

/* Return the protocol number. */
CSock& operator>> (CSock& sock, Protocol &pnum) {
    pnum = (Protocol)(sock.inbuf[2]);
    return sock;
}

/* This will send a message when << end; has been sent. */
void CSock::send_msg() {
    /* Calculate and store message length in packet */
    outptr++;				/* So the terminating zero is a
                                           part of the message. */
#ifdef SERIAL
    if (isatty(sockfd))
        outptr++;                       /* So terminating \r also counted*/
#endif /* SERIAL */
    outbuf[0] = (outptr - outbuf) % 128;
    outbuf[1] = ((outptr - outbuf) / 128);


    /* Send the buffer message, and flush the buffer.
       Ignore errors, as the select / read will pick these up. */
    write(sockfd, outbuf, (outptr - outbuf) - 1);
#ifdef SERIAL
    if (isatty(sockfd))
        write(sockfd, "\0\r", 2);
    else
#endif /* SERIAL */
    write(sockfd, "\0", 1);

#ifdef DEBUG_PACKETS
    log << "DEBUG:  Sent packet type=" << (int)outbuf[2] << "\tto="
         << sockfd << "\tlen=" << (int)(outptr - outbuf) << '\n';
#endif /* DEBUG_PACKETS */
}

/* Fill the input buffer
   returns -1 on error, or number of messages waiting in buffer,
   0 indicates no messages are ready. */
int CSock::fill_inbuf() {
    int retval;
    int msgsize;

    /* If there is a full message in the buffer, move it out */
    if ((inptr - inbuf) > 2) {
        msgsize = inbuf[0] + (inbuf[1] * 128);
        if (msgsize <= (inptr - inbuf) && inptr != inbuf) {
            mymemmove(inbuf, inbuf+msgsize, BUFFER_SIZE - msgsize);
            inptr -= msgsize;
        }
    }

    /* Check for a second message after the first, in which case we
       do not read from the socket, as this may block! */
    if ((inptr - inbuf) > 2) {
        msgsize = inbuf[0] + (inbuf[1] * 128);
        if (msgsize <= (inptr - inbuf)) {
            goto x;
        }
    }

    /* Read the new data in */
    retval = read(sockfd, inptr, BUFFER_SIZE - (inptr-inbuf));
    if (retval <= 0) {
        log << "ERROR: read socket failed" << '\n';
        return -1;
    } else {
        /* check to make sure we have at least the first two bytes */
        if (((inptr - inbuf) + retval) < 2)
            return 0;

        /* Set the end of the buffer, and the message size */
        inptr += retval;
        msgsize = inbuf[0] + (inbuf[1] * 128);

        /* If the message size is less than the data in the buffer,
           we can return a success. */
        if (msgsize <= (inptr - inbuf)) {
x:
            inread = inbuf + 3;

#ifdef DEBUG_PACKETS
    log << "DEBUG:  Recv packet type=" << (int)inbuf[2] << "\tfrom="
         << sockfd << "\tlen=" << (int)(msgsize) << '\n';
#endif
            /* Check for another message after this one. If there is
               another message and it is all in the buffer, return 2 */
            if ((msgsize + 2) <= (inptr - inbuf)) {
                int nextmsgsize;
                nextmsgsize = inptr[0] + (inptr[1] * 128);
                if (msgsize + nextmsgsize <= (inptr - inbuf))
                    return 2;
            }
            return 1;
        }
    }
    return 0;
}
