/* Interhack   Multi-player Rougelike game (based on Nethack)
   Copyright (C) 1996-2000  Damian Bentley

   See license.txt for details. If you do not have a copy of the license,
   please email bendchon@mehta.anu.edu.au */

#ifndef _string_h_
#define _string_h_

/* We have created our own string class mainly because during development
   we had many compatability problems between Linux, Solaris, and BSD.
   So the quickest solution was to write our own class. We use a string
   class rather than a char * because it saves problems when copying
   objects with pointers.

   Everything in strings is inline. They are used a fair bit, and so
   need to be relatively fast.

   How to use the string class:
       string s;
       string s("string");
       string s('c');
       string s(42);

       s = "hello";
       s += " world";
       s += ' ';
       s += 42;

       char c = s[4];

       cout << s;
       cout << s.getStr();
       cout << s.size();

       - Damian Bentley 15-10-99 */

#include "system.h"

inline void mystrncpy(char *dest, char *src, int len) {
    register int i;
    for (i = 0; i < len; i++)
        dest[i] = src[i];
}

inline void mymemmove(char *dest, char *src, int len) {
    char *buf = new char[len];
    mystrncpy(buf, src, len);
    mystrncpy(dest, buf, len);
    delete [] buf;
}

inline int mystrlen(char *str) {
    if (str == NULL) return 0;
    register int sindex = 0;
    while (str[sindex] != '\n' && str[sindex] != '\0') sindex++;
    return sindex;
}

inline void mystrcpy(char *dest, char *src) {
    mystrncpy(dest, src, mystrlen(src));
}

inline int mystrncmp(char *one, char *two, int n) {
    int i = 0;
    for (i = 0; i < n; i++)
        if (one[i] < two[i]) return (-i - 1);
        else if (one[i] > two[i]) return (i + 1);
    return 0;
}

inline int mystrcmp(char *one, char *two) {
    int n = mystrlen(one);
    if (mystrlen(two) < n) n = mystrlen(two);
    int i;
    for (i = 0; i < n; i++)
        if (one[i] < two[i]) return (-i - 1);
        else if (one[i] > two[i]) return (i + 1);
    if (mystrlen(one) > mystrlen(two)) return (-i - 1);
    else if (mystrlen(two) < mystrlen(one)) return (i + 1);
    return 0;
}

inline char * mystrstr(char *haystack, char *needle) {
    if (haystack == NULL || needle == NULL) return NULL;
    int index = 0, i;
    for (i = 0; i < mystrlen(haystack); i++) {
        if (index == mystrlen(needle)) return (haystack + i - index);
        if (haystack[i] == needle[index]) index++;
        else index = 0;
    }
    if (index == mystrlen(needle)) return (haystack + i - index);
    return NULL;
}

inline char *mystrdup(char *s) {
    if (s == NULL) return NULL;
    return new char[mystrlen(s)];
}

class string {
public:
    // Constructors and destructors
    string() { str = NULL; length = 0; };
    string(const string &s) {
        length = s.length;
        str = new char[length+1];
        if (length > 0)
            mystrcpy(str, s.str);
        str[length] = '\0'; };
    ~string() { if (str != NULL) delete [] str; str = NULL; length = 0; };

    // General functions
    int size() { return length; };
    char * getStr() { return str; };
    int getNumber() {
        int n = 0;
        bool negative = false;
        for (int i = 0; i < length; i++) {
            if ((str[i] < '0' || str[i] > '9') && str[i] != '-') break;
            if (str[i] != '-')
                n = (n*10) + (str[i] - '0');
            else
                negative = true;
        }
        if (negative) n = -n;
        return n; };

    // The funky functions
    string& operator= (const string &s) {
        if (str != NULL) delete [] str; str = NULL;
        length = s.length;
        str = new char[length+1];
        mystrcpy(str, s.str); str[length] = '\0';
        return *this; };
    string& operator+= (const string &s) {
        char * temp = new char[length + s.length + 1];
        mystrncpy(temp, str, length);
        mystrncpy(temp+length, s.str, s.length);
        temp[length + s.length] = '\0';
        delete [] str; str = temp;
        length += s.length;
        return *this; };
    char& operator[] (int num) {
        return str[num]; };
    int operator-- () { str[length--] = '\0'; return length; };

    friend int operator ==(string &s1, string &s2) {
        if (s1.length != s2.length) return 0;
        for (int z=0;z<s1.length;z++)
        {
            if (s1[z] != s2[z]) return 0;
        }
        return 1;}
    friend int operator ==(string &s1, char * s2) {
        if (s1.length != mystrlen(s2)) return 0;
        for (int z=0;z<s1.length;z++)
        {
            if (s1[z] != s2[z]) return 0;
        }
        return 1;}
    friend int operator !=(string &s1, string &s2) {
        if (s1 == s2) return 0;
        return 1; }

    // The conversion functions
    string(char *s) {
        if (s == NULL) return;
        length = mystrlen(s);
        str = new char[length+1];
        mystrcpy(str, s); str[length] = '\0'; };
    string(char c) {
        length = 1;
        str = new char[length+1];
        str[0] = c; str[1] = '\0'; };
    string(int n) {
        int ntemp = n;
        length = 0;
        char temp[8];
        if (n < 0) {
            temp[length++] = '-';
            n = -n;
        }
        for (int i = 100000; i > 0; i /= 10)
            if ((n / i) > 0 || length > 0) {
                temp[length] = ((n / i) + '0');
                n = n % i;
                length++;
            }
        if (ntemp == 0) temp[length++] = '0';
        str = new char[length+1];
        mystrncpy(str, temp, length); str[length] = '\0'; };
    // Translate to upper case
    void upper() {
        for (int i = 0; i < length; i++)
            if (str[i] >= 'a' && str[i] <= 'z')
                str[i] = str[i] + ('A' - 'a'); }
    // Translate to lower case
    void lower() {
        for (int i = 0; i < length; i++)
            if (str[i] >= 'A' && str[i] <= 'Z')
                str[i] = str[i] - ('A' - 'a'); }
    // Confirm it is a valid str (ascii 33..126)
    void validStr() {
        for (int i = 0; i < length; i++)
            if (str[i] < 33 || str[i] > 126)
                str[i] = ' '; }

    // A sorting function (used in lists of objects)
    // Sorts letters into alphabetical order, then reduces the list
    string sort() {
        // Sort the letters (no need for a fancy algorithm)
        for (int i = 0; i < length; i++)
            for (int j = i; j < length; j++)
                if (str[i] > str[j]) {
                    char t = str[i];
                    str[i] = str[j];
                    str[j] = t;
                }
        // Reduce the list. For 3 or more consecutive letters: a-d
        string t;
        int c = 0;
        for (int i = 0; i < length; i++)
            if (i > 0 && i < length-1 && str[i] == (str[i-1])+1) c++;
            else if (c == 0) t += str[i];
            else {
                if (i == length-1 && c == 0) t += str[i];
                else if (i == length-1) { c = 0; t += '-'; t += str[i]; }
                else { c = 0; t += '-'; t += str[i-1]; } }
        return t; };

    // The search functions... (none needed yet)
    bool contains(char c) {
        for (int i = 0; i < length; i++)
            if (str[i] == c) return true;
        return false;
    }
private:
    int length;
    char * str;

protected:
};

#endif
