/* Interhack   Multi-player Rougelike game (based on Nethack)
   Copyright (C) 1996-2000  Damian Bentley

   See license.txt for details. If you do not have a copy of the license,
   please email bendchon@mehta.anu.edu.au */

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>

#ifdef SERIAL
#include <termios.h>
#include <fcntl.h>
#include <string.h>
#endif /* SERIAL */

#include "tcpip.h"
#include "log.h"
#include "plist.h"

#include <curses.h>

#ifndef MOUSE
#define MOUSE " "
#endif

/* Check at least one of the connection types are in use */
#ifndef TCPIP
  #ifndef SERIAL
    #error "You need to define TCPIP or SERIAL in Makefile.all!"
  #endif /* SERIAL */
#endif /* TCPIP */

extern int incomming_data();

/* Setup everything for a connection */
CConnection::CConnection(int (*n_conn)(int, int, char *),
    int (*d_in)(int), int (*c_conn)(int)) {
    int x;

    maxfd = -1;
    num_of_connections = 0;
    sockfd = csockfd = -1;
    serialfd0 = serialfd1 = -1;
    oldserialfd0 = oldserialfd1 = -1;
    FD_ZERO(&read_fds);
    for (x = 0; x < MAXPLAYERS; x++) {
        connection[x].fds = -1;
        ih_timerclear(&(connection[x].lastdata));
    }
    new_conn = n_conn;
    data_in = d_in;
    close_conn = c_conn;
    demon = false;
}

/* Destruct everything */
CConnection::~CConnection() {
}

/* Open SERVER_PORT, bind it and listen to it for connections and data. */
int CConnection::c_open(int port) {
#ifdef TCPIP
    struct sockaddr_in server;
#endif /* TCPIP */
    int x;

    if (c_sigpipe_ignore() < 0) return -1;

#ifdef TCPIP
    log << "TCPIP: listening on port " << port << '\n';

    if ((port < PORTMIN) || (port > PORTMAX)) {
        port = SERVER_PORT;
        if ((port < PORTMIN) || (port > PORTMAX))
            return -1;
    }

    if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
        return -1;

    server.sin_family = AF_INET;
    server.sin_addr.s_addr = INADDR_ANY;
    server.sin_port = htons(port);

    if (bind(sockfd, (struct sockaddr *)&server, sizeof(server)) < 0)
        return -1;
    if (listen(sockfd, MAXBACKLOG) < 0)
        return -1;
#endif /* TCPIP */

#ifdef SERIAL
  #ifdef MOUSE
    log << "SERIAL: mouse is on " << MOUSE << '\n';
    if (mystrcmp(SERIAL0, MOUSE) != 0)
  #endif /* MOUSE */
        oldserialfd0 = serialfd0 = open_serial(SERIAL0);
  #ifdef MOUSE
    if (mystrcmp(SERIAL1, MOUSE) != 0)
  #endif /* MOUSE */
        oldserialfd1 = serialfd1 = open_serial(SERIAL1);
  #ifndef TCPIP
    if (serialfd0 < 0 && serialfd1 < 0) {
        log << "ERROR: could not open a serial port, tcpip not in use" << '\n';
        return -1;
    }
  #endif /* TCPIP */
    if (serialfd0 < 0 && strcmp(MOUSE, SERIAL0) != 0) {
        log << "ERROR: could not open " << SERIAL0 << '\n';
        log << "       (" << strerror(errno) << ')' << '\n';
    }
    if (serialfd1 < 0 && strcmp(MOUSE, SERIAL1) != 0) {
        log << "ERROR: could not open " << SERIAL1 << '\n';
        log << "       (" << strerror(errno) << ')' << '\n';
    }
#endif /* SERIAL */

    for (x = 0; x < MAXPLAYERS; x++) {
        connection[x].fds = -1;
        ih_timerclear(&(connection[x].lastdata));
    }

    return 0;
}

#ifdef SERIAL
/* Prepare a serial port sp, and return a fd to it */
int CConnection::open_serial(char *sp) {
    int fd = -1;
    struct termios oldtio,newtio;

    if (sp == NULL) {
        log << "ERROR: open_serial(NULL)!" << '\n';
        return -1;
    }
    log << "SERIAL: opening=" << sp << '\n';

    /* open the device to be non-blocking (read will return immediatly) */
    fd = open(sp, O_RDWR | O_NOCTTY | O_NONBLOCK);
    if (fd < 0) return -1;
    log << "SERIAL: fd=" << fd << '\n';

    /* allow the process to receive SIGIO */
    fcntl(fd, F_SETOWN, getpid());
    /* Make the file descriptor asynchronous (the manual page says only
       O_APPEND and O_NONBLOCK, will work with F_SETFL...) */
    fcntl(fd, F_SETFL, /*FASYNC*/ FNDELAY);

    tcgetattr(fd,&oldtio); /* save current port settings */
    /* set new port settings for canonical input processing */
    newtio.c_cflag = BAUDRATE | CRTSCTS | CS8 | CLOCAL | CREAD;
    newtio.c_iflag = IGNPAR | ICRNL;
    newtio.c_oflag = 0;
    newtio.c_lflag = ICANON;
    newtio.c_cc[VMIN]=0;
    newtio.c_cc[VTIME]=0;
    tcflush(fd, TCIFLUSH);
    tcsetattr(fd,TCSANOW,&newtio);

    return fd;
}
#endif

#ifdef TCPIP
/* Deals with new connections to the server. */
int CConnection::c_new_tcpipconn() {
    struct sockaddr_in client;
    int x, newfd, client_len;

    client_len = sizeof(client);
    if ((newfd = accept(sockfd, (struct sockaddr *)&client, &client_len)) == -1) {
        log << "WARNING: Could not accept connection. "
             << "You may need to increase MAXBACKLOG." << '\n';
        return -1;
    }
    if (num_of_connections == MAXPLAYERS) {
        log << "WARNING: Too many players trying to connect. "
             << "You may need to increase MAXPLAYERS." << '\n';
        close(newfd);
        return 0;
    }
    if (newfd >= maxfd) maxfd = newfd + 1;
    if (csockfd >= maxfd) maxfd = csockfd + 1;
    FD_SET(newfd, &read_fds);
    for (x = 0; x < MAXPLAYERS; x++)
        if (connection[x].fds == -1) {
            connection[x].fds = newfd;
            mystrcpy(connection[x].clientname, inet_ntoa(client.sin_addr));
            break;
        }
    if (x == MAXPLAYERS) {
        log << "ERROR: CConnection:c_new_conn() - could not find "
             << "empty slot, yet should be one spare." << '\n';
        close(newfd);
        return 0;
    }
    num_of_connections++;
    new_conn(x, newfd, connection[x].clientname);

    return 0;
}
#endif /* TCPIP */

#ifdef SERIAL
/* Deals with serial connections and data to the server. */
int CConnection::c_new_serialconn(int fd) {
    int newfd, x;

    log << "SERIAL: new connection and data being received" << '\n';

    if (fd == 0)
        newfd = serialfd0;
    else if (fd == 1)
        newfd = serialfd1;
    else {
        log << "ERROR: c_new_serialconn, unknown fd" << '\n';
        return -1;
    }

    if (num_of_connections == MAXPLAYERS) {
        log << "WARNING: Too many players trying to connect. "
             << "You may need to increase MAXPLAYERS." << '\n';
        return 0;
    }
    FD_SET(newfd, &read_fds);
    for (x = 0; x < MAXPLAYERS; x++)
        if (connection[x].fds == -1) {
            connection[x].fds = newfd;
            if (fd == 0)
                mystrcpy(connection[x].clientname, SERIAL0);
            else
                mystrcpy(connection[x].clientname, SERIAL1);
            break;
        }
    if (x == MAXPLAYERS) {
        log << "ERROR: CConnection:c_new_conn() - could not find "
             << "empty slot, yet should be one spare." << '\n';
        return 0;
    }
    num_of_connections++;
    new_conn(x, newfd, connection[x].clientname);
    /* This is hack to make sure no more connections from this
       serial port until it is disconnected */
    if (fd == 0) serialfd0 = -1;
    else         serialfd1 = -1;
    return 0;
}
#endif /* SERIAL */

/* Deals with connections which have been closed. */
int CConnection::c_close_conn(int id) {
    int x;

    if (close_conn != NULL)
        close_conn(id);

    FD_CLR(connection[id].fds, &read_fds);
#ifdef SERIAL
    /* If the serial connection has closed, we actually want to keep it
       open - otherwise it will be lost forever */
    if (oldserialfd0 == connection[id].fds || oldserialfd1 == connection[id].fds) {
        if (oldserialfd0 == connection[id].fds)
            serialfd0 = oldserialfd0;
        else
            serialfd1 = oldserialfd1;
    } else
#else /* SERIAL / TCPIP */
    {        /* Close any tcpip connections */
    close(connection[id].fds);
    connection[id].fds = -1;
    }
#endif /* TCPIP */
    ih_timerclear(&(connection[id].lastdata));
    num_of_connections--;
    maxfd = -1;
    for (x = 0; x < MAXPLAYERS; x++)
        if (connection[x].fds >= maxfd) maxfd = connection[x].fds + 1;
#ifdef TCPIP
    if (sockfd >= maxfd) maxfd = sockfd + 1;
    if (csockfd >= maxfd) maxfd = csockfd + 1;
#endif /* TCPIP */
#ifdef SERIAL
    if (serialfd0 >= maxfd) maxfd = serialfd0 + 1;
    if (serialfd1 >= maxfd) maxfd = serialfd1 + 1;
#endif /* SERIAL */
    return 0;
}

#define addfd(i)			\
    if (i >= 0)				\
        FD_SET(i, &read_fds);		\
    if (i >= maxfd) maxfd = i+1;

/* Wait for incoming connections and data, and deal with them. */
int CConnection::c_listen() {
    int retval;
    fd_set temp_fds;
#ifdef SERVER
    Timeval deadtime = DEADTIME;
    int x;
#endif

    /* Setup all the file descriptors, and maximum fd */
    maxfd = -1;
    FD_ZERO(&read_fds);
    if (!demon)                   /* If not in demon mode */
        addfd(0);                     /* Keyboard input */
#ifdef TCPIP
    addfd(sockfd);                    /* Server socket */
    addfd(csockfd);                   /* Client socket */
#endif /* TCPIP */
#ifdef SERIAL
    addfd(serialfd0);                 /* Serial port 0 */
    addfd(serialfd1);                 /* Serial port 1 */
#endif /* SERIAL */
#ifdef SERVER
    for (x = 0; x < MAXPLAYERS; x++) {
        addfd(connection[x].fds);     /* Player connections */
    }
#endif /* SERVER */


    while (1) {
        /* Wait for new connections or incoming data. */
        temp_fds = read_fds;
#ifdef SERVER
        Timeval * t = players.sleepTime();
#else
        Timeval * t = NULL;
#endif
        if ((retval = select(maxfd, &temp_fds, NULL, NULL, t)) == -1)
            continue;

#ifdef SERVER
        /* Convert the time returned into time used in call */
        Timeval sleept = {5, 0};
        ih_timersub(&sleept, *t);

        /* Make sure there are no dead connections */
        for (x = 0; x < MAXPLAYERS; x++)
            if (connection[x].fds != -1) {
                ih_timeradd(&connection[x].lastdata,sleept);
                if (ih_timercmp(connection[x].lastdata, deadtime, >=)) {
                    log << "WARNING: dead connection " << x << '\n';
                    log << "TODO: save player " << x << '\n';
                    c_close_conn(x);
                }
            }

        /* Check for incoming data, and deal with it. */
        for (x = 0; (x < MAXPLAYERS) && (retval > 0); x++) {
            if (connection[x].fds != -1 && FD_ISSET(connection[x].fds, &temp_fds)) {
                retval--;
                if (data_in(x) < 0) c_close_conn(x);
                ih_timerclear(&(connection[x].lastdata)); /* Dead conn timer */
            }
        }
#endif /* SERVER */

        /* Check for incoming data */
        if (csockfd > 0 && FD_ISSET(csockfd, &temp_fds))
            if (incomming_data() < 0) return -1; /* special case */
        /* Check key has not been pressed on server */
        if (!demon && FD_ISSET(0, &temp_fds))
            return getch();
#ifdef SERVER
  #ifdef TCPIP
        /* Accept a new connection if there is one. */
        if (FD_ISSET(sockfd, &temp_fds))
            if (c_new_tcpipconn() == -1)
                return -1;
  #endif /* TCPIP */
  #ifdef SERIAL
        /* Accept a new serial connection if there is one. */
        if (FD_ISSET(serialfd0, &temp_fds))
            if (c_new_serialconn(0) == -1)
                return -1;
        if (FD_ISSET(serialfd1, &temp_fds))
            if (c_new_serialconn(1) == -1)
                return -1;
  #endif /* SERIAL */

        // Do any forced moves
        players.forcedMoves();

        /* Make any moves that are banked up */
        players.makeMoves();
#endif /* SERVER */
    }

    return 0;
}

/* Connect to SERVER_NAME:PORT.
   If servername == modem or null-modem, and serial connection is enabled,
   use it. */
int CConnection::c_connect(char *servername, int port) {
#ifdef TCPIP
    struct sockaddr_in server;
    struct hostent *hp;
    int retval;
#endif /* TCPIP */
    csockfd = 0;

#ifdef SERIAL
    /* Is the client using a null-modem connection? */
    if (mystrcmp(servername, "null-modem") == 0) {
        log << "SERIAL: trying serial null-modem connection" << '\n';
        /* Use the first available line, unless used by mouse */
        if (mystrcmp(SERIAL0, MOUSE) != 0)
            serialfd0 = csockfd = open_serial(SERIAL0);
        else if (mystrcmp(SERIAL1, MOUSE) != 0)
            serialfd1 = csockfd = open_serial(SERIAL1);
        else goto x;
        if (csockfd < 0) {
            log << "ERROR: could not open serial port for client to connect" << '\n';
            return -1;
        }
        return csockfd;
    } else if (mystrcmp(servername, "modem") == 0) {
        log << "SERIAL: trying serial modem connection" << '\n';
        log << "TODO: modem connections to the server" << '\n';
        return -1;
    }
x:
#endif /* SERIAL */

#ifdef TCPIP
    log << "TCPIP: falling back to tcpip connection" << '\n';
    if ((port < PORTMIN) || (port > PORTMAX)) {
        port = SERVER_PORT;
        if ((port < PORTMIN) || (port > PORTMAX))
            return -1;
    }
    if (servername == NULL) {
        servername = new char[mystrlen(SERVER_NAME)];
        mystrcpy(servername, SERVER_NAME);
    }

    if (c_sigpipe_ignore() != 0) return -1;
    if ((hp = gethostbyname(servername)) == NULL)
        return -1;
    if ((csockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
        return -1;

    memcpy((char *)&server.sin_addr, hp->h_addr_list[0], hp->h_length);

    server.sin_port = htons(port);
    server.sin_family = AF_INET;

    while (((retval = connect(csockfd, (struct sockaddr *)&server, sizeof(server))) == -1) &&
        (errno == EINTR))
        ;
    if (retval == -1) {
        close(csockfd);
        return -1;
    }
#endif /* TCPIP */

    return csockfd;
}

/* Close the main file descriptor. */
int CConnection::c_close() {
    int x;

#ifdef TCPIP
    if (csockfd > 0) {
        x = close(csockfd);
        csockfd = -1;
        return x;
    }
#endif /* TCPIP */

    for (x = 0; x < MAXPLAYERS; x++)
        if (connection[x].fds != -1)
            c_close_conn(x);

#ifdef TCPIP
    x = close(sockfd);
    sockfd = -1;
#endif /* TCPIP */
    return x;
}

/* Catch the sigpipe signal and send it to c_sigpipe(). */
int CConnection::c_sigpipe_ignore() {
#ifndef HAS_SIGIGNORE
    struct sigaction act;

    act.sa_handler = SIG_IGN;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
    if (sigaction(SIGPIPE, &act, NULL) < 0)
        return -1;
  #ifdef SERIAL
    act.sa_handler = SIG_IGN;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
    if (sigaction(SIGIO, &act, NULL) < 0)
        return -1;
  #endif /* SERIAL */
#else /* HAS_SIGIGNORE */
    sigignore(SIGPIPE);
    sigignore(SIGIO);
#endif /* HAS_SIGIGNORE */
    return 0;
}
